import logging
import os
import random

import matplotlib.pyplot as plt
import foldedleastsquares
from lcbuilder.constants import MISSION_TESS, MISSION_ID_TESS, MISSION_KEPLER
from lcbuilder.helper import LcbuilderHelper
from lcbuilder.star.HabitabilityCalculator import HabitabilityCalculator
from numpy.random import default_rng
import astropy.units as u
from exoml.ete6.ete6_generator import Ete6ModelGenerator
from sklearn.utils import shuffle
import numpy as np
import pandas as pd


class IatsonPlanetModelGenerator(Ete6ModelGenerator):
    def __init__(self, injected_objects_df, lcs_dir, batch_size, input_sizes, type_to_label, zero_epsilon=1e-7,
                 measurements_per_point=1, plot_inputs=False, fixed_target_id=None, store_arrays=False,
                 from_arrays=True, shuffle_batch=True, validation_objects_df=None, mask_previous_signals=False,
                 flag_incomplete=False, throw_nan_exception=False, offset_mode='mean', explain=False):
        super().__init__(zero_epsilon, shuffle_batch=shuffle_batch, flag_incomplete=flag_incomplete, throw_nan_exception=throw_nan_exception)
        self.injected_objects_df = injected_objects_df
        self.injected_objects_df = injected_objects_df
        self.EXPLAIN_PERIOD_VALUE = [self.zero_epsilon, 0.5, 1 - self.zero_epsilon]
        self.EXPLAIN_DURATION_VALUE = [self.zero_epsilon, 0.5, 1 - self.zero_epsilon]
        self.EXPLAIN_RADIUS_VALUE = [self.zero_epsilon, 0.5, 1 - self.zero_epsilon]
        self.EXPLAIN_GTCN_VALUE = [self.zero_epsilon, 0.5, 1 - self.zero_epsilon]
        self.EXPLAIN_GTR_VALUE = [self.zero_epsilon, 0.5, 1 - self.zero_epsilon]
        self.EXPLAIN_OFFSET_VALUE = [self.zero_epsilon, 0.5, 1 - self.zero_epsilon]
        self.EXPLAIN_BFAP_VALUE = [self.zero_epsilon, 0.5, 1 - self.zero_epsilon]
        self.EXPLAIN_ODDEVEN_VALUE = [self.zero_epsilon, 0.5, 1 - self.zero_epsilon]
        self.EXPLAIN_TEMP_VALUE = [self.zero_epsilon, 0.5, 1 - self.zero_epsilon]
        self.EXPLAIN_ALBEDO_VALUE = [self.zero_epsilon, 0.5, 1 - self.zero_epsilon]
        self.EXPLAIN_STARTEFF_VALUE = [self.zero_epsilon, 0.5, 1 - self.zero_epsilon]
        self.EXPLAIN_STARRAD_VALUE = [self.zero_epsilon, 0.5, 1 - self.zero_epsilon]
        self.kics_lcs_dir = lcs_dir + '/q1_q17/lcs/'
        self.ete6_lcs_dir = lcs_dir + '/ete6/lcs/'
        self.tess_lcs_dir = lcs_dir + '/tess/lcs/'
        self.batch_size = batch_size
        self.input_sizes = input_sizes
        self.random_number_generator = default_rng()
        self.measurements_per_point = measurements_per_point
        self.type_to_label = type_to_label
        self.plot_inputs = plot_inputs
        self.fixed_target_id = fixed_target_id
        self.store_arrays = store_arrays
        self.from_arrays = from_arrays
        self.validation_objects_df = validation_objects_df
        self.mask_previous_signals = mask_previous_signals
        self.offset_mode = offset_mode
        self.explain = explain
        if self.explain:
            final_df = pd.DataFrame(columns=['object_id', 'period', 'duration(h)', 'depth_primary', 'radius(earth)', 'type',
                                             'EXPLAIN_PERIOD_VALUE', 'EXPLAIN_DURATION_VALUE', 'EXPLAIN_RADIUS_VALUE', 'EXPLAIN_GTCN_VALUE',
                                             'EXPLAIN_GTR_VALUE', 'EXPLAIN_OFFSET_VALUE', 'EXPLAIN_BFAP_VALUE',
                                             'EXPLAIN_ODDEVEN_VALUE', 'EXPLAIN_TEMP_VALUE', 'EXPLAIN_ALBEDO_VALUE',
                                             'EXPLAIN_STARTEFF_VALUE', 'EXPLAIN_STARRAD_VALUE',
                                             'EXPLAIN_GLOBAL_VIEW', 'EXPLAIN_MAIN_VIEW', 'EXPLAIN_SECONDARY_VIEW',
                                             'EXPLAIN_ODD_VIEW', 'EXPLAIN_EVEN_VIEW', 'EXPLAIN_CENTROIDS_RA_VIEW',
                                             'EXPLAIN_CENTROIDS_DEC_VIEW','EXPLAIN_OG_VIEW'])
            for index, row in self.injected_objects_df.iterrows():
                row_copy_df = row.copy().to_frame().T
                final_df = pd.concat([final_df, row_copy_df], ignore_index=True)
                row_copy_df = row.copy().to_frame().T
                row_copy_df['EXPLAIN_GLOBAL_VIEW'] = True
                row_copy_df['object_id'] = row_copy_df['object_id'] + '_global'
                final_df = pd.concat([final_df, row_copy_df], ignore_index=True)
                row_copy_df = row.copy().to_frame().T
                row_copy_df['EXPLAIN_MAIN_VIEW'] = True
                row_copy_df['object_id'] = row_copy_df['object_id'] + '_main'
                final_df = pd.concat([final_df, row_copy_df], ignore_index=True)
                row_copy_df = row.copy().to_frame().T
                row_copy_df['EXPLAIN_SECONDARY_VIEW'] = True
                row_copy_df['object_id'] = row_copy_df['object_id'] + '_secondary'
                final_df = pd.concat([final_df, row_copy_df], ignore_index=True)
                row_copy_df = row.copy().to_frame().T
                row_copy_df['EXPLAIN_ODD_VIEW'] = True
                row_copy_df['object_id'] = row_copy_df['object_id'] + '_odd'
                final_df = pd.concat([final_df, row_copy_df], ignore_index=True)
                row_copy_df = row.copy().to_frame().T
                row_copy_df['EXPLAIN_EVEN_VIEW'] = True
                row_copy_df['object_id'] = row_copy_df['object_id'] + '_even'
                final_df = pd.concat([final_df, row_copy_df], ignore_index=True)
                row_copy_df = row.copy().to_frame().T
                row_copy_df['EXPLAIN_CENTROIDS_RA_VIEW'] = True
                row_copy_df['object_id'] = row_copy_df['object_id'] + '_ra'
                final_df = pd.concat([final_df, row_copy_df], ignore_index=True)
                row_copy_df = row.copy().to_frame().T
                row_copy_df['EXPLAIN_CENTROIDS_DEC_VIEW'] = True
                row_copy_df['object_id'] = row_copy_df['object_id'] + '_dec'
                final_df = pd.concat([final_df, row_copy_df], ignore_index=True)
                row_copy_df = row.copy().to_frame().T
                row_copy_df['EXPLAIN_OG_VIEW'] = True
                row_copy_df['object_id'] = row_copy_df['object_id'] + '_og'
                final_df = pd.concat([final_df, row_copy_df], ignore_index=True)
                for value in self.EXPLAIN_PERIOD_VALUE:
                    row_copy_df = row.copy().to_frame().T
                    row_copy_df['EXPLAIN_PERIOD_VALUE'] = value
                    row_copy_df['object_id'] = row_copy_df['object_id'] + f'_period_{value}'
                    final_df = pd.concat([final_df, row_copy_df], ignore_index=True)
                for value in self.EXPLAIN_DURATION_VALUE:
                    row_copy_df = row.copy().to_frame().T
                    row_copy_df['EXPLAIN_DURATION_VALUE'] = value
                    row_copy_df['object_id'] = row_copy_df['object_id'] + f'_duration_{value}'
                    final_df = pd.concat([final_df, row_copy_df], ignore_index=True)
                for value in self.EXPLAIN_RADIUS_VALUE:
                    row_copy_df = row.copy().to_frame().T
                    row_copy_df['EXPLAIN_RADIUS_VALUE'] = value
                    row_copy_df['object_id'] = row_copy_df['object_id'] + f'_radius_{value}'
                    final_df = pd.concat([final_df, row_copy_df], ignore_index=True)
                for value in self.EXPLAIN_GTCN_VALUE:
                    row_copy_df = row.copy().to_frame().T
                    row_copy_df['EXPLAIN_GTCN_VALUE'] = value
                    row_copy_df['object_id'] = row_copy_df['object_id'] + f'_gtcn_{value}'
                    final_df = pd.concat([final_df, row_copy_df], ignore_index=True)
                for value in self.EXPLAIN_GTR_VALUE:
                    row_copy_df = row.copy().to_frame().T
                    row_copy_df['EXPLAIN_GTR_VALUE'] = value
                    row_copy_df['object_id'] = row_copy_df['object_id'] + f'_gtr_{value}'
                    final_df = pd.concat([final_df, row_copy_df], ignore_index=True)
                for value in self.EXPLAIN_OFFSET_VALUE:
                    row_copy_df = row.copy().to_frame().T
                    row_copy_df['EXPLAIN_OFFSET_VALUE'] = value
                    row_copy_df['object_id'] = row_copy_df['object_id'] + f'_offset_{value}'
                    final_df = pd.concat([final_df, row_copy_df], ignore_index=True)
                for value in self.EXPLAIN_BFAP_VALUE:
                    row_copy_df = row.copy().to_frame().T
                    row_copy_df['EXPLAIN_BFAP_VALUE'] = value
                    row_copy_df['object_id'] = row_copy_df['object_id'] + f'_bfap_{value}'
                    final_df = pd.concat([final_df, row_copy_df], ignore_index=True)
                for value in self.EXPLAIN_ODDEVEN_VALUE:
                    row_copy_df = row.copy().to_frame().T
                    row_copy_df['EXPLAIN_ODDEVEN_VALUE'] = value
                    row_copy_df['object_id'] = row_copy_df['object_id'] + f'_oddeven_{value}'
                    final_df = pd.concat([final_df, row_copy_df], ignore_index=True)
                for value in self.EXPLAIN_TEMP_VALUE:
                    row_copy_df = row.copy().to_frame().T
                    row_copy_df['EXPLAIN_TEMP_VALUE'] = value
                    row_copy_df['object_id'] = row_copy_df['object_id'] + f'_temp_{value}'
                    final_df = pd.concat([final_df, row_copy_df], ignore_index=True)
                for value in self.EXPLAIN_ALBEDO_VALUE:
                    row_copy_df = row.copy().to_frame().T
                    row_copy_df['EXPLAIN_ALBEDO_VALUE'] = value
                    row_copy_df['object_id'] = row_copy_df['object_id'] + f'_albedo_{value}'
                    final_df = pd.concat([final_df, row_copy_df], ignore_index=True)
                for value in self.EXPLAIN_STARTEFF_VALUE:
                    row_copy_df = row.copy().to_frame().T
                    row_copy_df['EXPLAIN_STARTEFF_VALUE'] = value
                    row_copy_df['object_id'] = row_copy_df['object_id'] + f'_starteff_{value}'
                    final_df = pd.concat([final_df, row_copy_df], ignore_index=True)
                for value in self.EXPLAIN_STARRAD_VALUE:
                    row_copy_df = row.copy().to_frame().T
                    row_copy_df['EXPLAIN_STARRAD_VALUE'] = value
                    row_copy_df['object_id'] = row_copy_df['object_id'] + f'_starrad_{value}'
                    final_df = pd.concat([final_df, row_copy_df], ignore_index=True)
            self.injected_objects_df = final_df

    def __len__(self):
        return (np.ceil(len(self.injected_objects_df) / float(self.batch_size))).astype(int)

    def class_weights(self):
        return {0: 1, 1: 1}

    def _plot_df(self, df, type, scenario):
        fig, axs = plt.subplots(2, 3, figsize=(12, 6), constrained_layout=True)
        axs[0][0].scatter(df['#time'], df['flux'])
        axs[0][1].scatter(df['#time'], df['centroid_x'])
        axs[0][2].scatter(df['#time'], df['centroid_y'])
        axs[1][0].scatter(df['#time'], df['bck_flux'])
        axs[1][1].scatter(df['#time'], df['motion_x'])
        axs[1][2].scatter(df['#time'], df['motion_y'])
        plt.title(type + " " + scenario)
        plt.show()
        plt.clf()
        plt.close()

    def _plot_numeric_values(self, target_id, scalar_values, star_array):
        if self.plot_inputs:
            logging.info(f"{target_id} raw values: {scalar_values[0] * 1200} d, {scalar_values[1] * 15} d, {scalar_values[2] * 1000} ppt, {scalar_values[3] * 300} Re,"
                         f" {scalar_values[4] * 20} gtcn, {scalar_values[5]} gtr, {scalar_values[6]} arcsec, {scalar_values[7]} arcsec_err, {scalar_values[8]} fap,"
                         f" {scalar_values[9]} oef, {star_array[2] * 60000} K, {star_array[5] * 3} Rsun, {star_array[6] * 3} M_sun")
            logging.info(f"{target_id} normalized values: {scalar_values[0]} d, {scalar_values[1]} d, {scalar_values[2]} ppt, {scalar_values[3]} Re,"
                         f" {scalar_values[4]} gtcn, {scalar_values[5]} gtr, {scalar_values[6]} arcsec, {scalar_values[7]} arcsec_err, {scalar_values[8]} fap,"
                         f" {scalar_values[9]} oef, {star_array[2]} K, {star_array[5]} Rsun, {star_array[6]} M_sun")


    def _plot_input(self, input_array, type, scenario):
        if self.plot_inputs:
            fig, axs = plt.subplots(1, 1, figsize=(24, 12), constrained_layout=True)
            current_array_mask = np.argwhere((~np.isnan(input_array)) & (input_array > self.zero_epsilon)).flatten()
            current_array = input_array[current_array_mask]
            axs.scatter(np.arange(0, len(current_array)), current_array)
            axs.set_title(type + '_' + scenario)
            fig.suptitle(type + " " + scenario)
            plt.show()
            plt.clf()
            plt.close()

    def mask_other_signals(self, data_df, target_row, time_key):
        other_signals_df = self.injected_objects_df[(self.injected_objects_df['object_id'] == target_row['object_id']) &
                                                    (self.injected_objects_df['period'] != target_row['period'])]
        if self.validation_objects_df is not None:
            other_signals_val_df = self.validation_objects_df[(self.validation_objects_df['object_id'] == target_row['object_id']) &
                                                              (self.validation_objects_df['period'] != target_row['period'])]
            other_signals_df = pd.concat([other_signals_df, other_signals_val_df])
        for index, other_signal_row in other_signals_df.iterrows():
            mask_signal = True
            if 'KIC' in target_row['object_id']:
                other_koi_tce_plnt_num = 999
                other_tce_plnt_num = 999
                target_koi_tce_plnt_num = 999
                target_tce_plnt_num = 999
                if 'koi_tce_plnt_num_y' in other_signal_row and not np.isnan(other_signal_row['koi_tce_plnt_num_y']):
                    other_koi_tce_plnt_num = other_signal_row['koi_tce_plnt_num_y']
                elif 'koi_tce_plnt_num_x' in other_signal_row and not np.isnan(other_signal_row['koi_tce_plnt_num_x']):
                    other_koi_tce_plnt_num = other_signal_row['koi_tce_plnt_num_x']
                elif 'koi_tce_plnt_num' in other_signal_row and not np.isnan(other_signal_row['koi_tce_plnt_num']):
                    other_koi_tce_plnt_num = other_signal_row['koi_tce_plnt_num']
                if 'koi_tce_plnt_num_y' in target_row and not np.isnan(target_row['koi_tce_plnt_num_y']):
                    target_koi_tce_plnt_num = target_row['koi_tce_plnt_num_y']
                elif 'koi_tce_plnt_num_x' in target_row and not np.isnan(target_row['koi_tce_plnt_num_x']):
                    target_koi_tce_plnt_num = target_row['koi_tce_plnt_num_x']
                elif 'koi_tce_plnt_num' in target_row and not np.isnan(target_row['koi_tce_plnt_num']):
                    target_koi_tce_plnt_num = target_row['koi_tce_plnt_num']
                if 'tce_plnt_num_y' in other_signal_row and not np.isnan(other_signal_row['tce_plnt_num_y']):
                    other_tce_plnt_num = other_signal_row['tce_plnt_num_y']
                elif 'tce_plnt_num_x' in other_signal_row and not np.isnan(other_signal_row['tce_plnt_num_x']):
                    other_tce_plnt_num = other_signal_row['tce_plnt_num_x']
                elif 'tce_plnt_num' in other_signal_row and not np.isnan(other_signal_row['tce_plnt_num']):
                    other_tce_plnt_num = other_signal_row['tce_plnt_num']
                if 'tce_plnt_num_y' in target_row and not np.isnan(target_row['tce_plnt_num_y']):
                    target_tce_plnt_num = target_row['tce_plnt_num_y']
                elif 'tce_plnt_num_x' in target_row and not np.isnan(target_row['tce_plnt_num_x']):
                    target_tce_plnt_num = target_row['tce_plnt_num_x']
                elif 'tce_plnt_num' in target_row and not np.isnan(target_row['tce_plnt_num']):
                    target_tce_plnt_num = target_row['tce_plnt_num']
                mask_signal = (not np.isnan(target_koi_tce_plnt_num) and not np.isnan(other_koi_tce_plnt_num) and \
                    target_koi_tce_plnt_num > other_koi_tce_plnt_num) or \
                              (not (np.isnan(target_tce_plnt_num) & np.isnan(other_tce_plnt_num)) and \
                    target_tce_plnt_num > other_tce_plnt_num)
            if mask_signal:
                mask = foldedleastsquares.transit_mask(data_df[time_key].to_numpy(), other_signal_row['period'],
                                                       2 * other_signal_row['duration(h)'] / 24, other_signal_row['epoch'])
                data_df = data_df[~mask]
        return data_df


    def plot_single_data(self, lc_df, target_row):
        fig, axs = plt.subplots(1, 1, figsize=(8, 4), constrained_layout=True)
        axs.scatter(lc_df['#time'], lc_df['flux_0'])
        axs.set_ylim(0.5 - target_row['tce_depth'] / 0.5e6, 0.505)
        plt.show()
        plt.clf()
        plt.close()

    def __getitem__(self, idx):
        max_index = (idx + 1) * self.batch_size
        max_index = len(self.injected_objects_df) if max_index > len(self.injected_objects_df) else max_index
        target_indexes = np.arange(idx * self.batch_size, max_index, 1)
        injected_objects_df = self.injected_objects_df.iloc[target_indexes]
        if self.shuffle_batch:
            injected_objects_df = shuffle(injected_objects_df)
        star_array = np.empty((len(target_indexes), self.input_sizes[0], 1))
        #[period, planet radius, number of transits, ratio of good transits, transit depth, transit_offset_pos - transit_offset_err]
        scalar_values = np.empty((len(target_indexes), 12, 1))
        global_flux_array = np.empty((len(target_indexes), self.input_sizes[2], self.measurements_per_point))
        folded_flux_even_array = np.empty((len(target_indexes), self.input_sizes[3], self.measurements_per_point))
        folded_flux_odd_array = np.empty((len(target_indexes), self.input_sizes[3], self.measurements_per_point))
        folded_flux_even_subhar_array = np.empty((len(target_indexes), self.input_sizes[3], self.measurements_per_point))
        folded_flux_odd_subhar_array = np.empty((len(target_indexes), self.input_sizes[3], self.measurements_per_point))
        folded_flux_even_har_array = np.empty((len(target_indexes), self.input_sizes[3], self.measurements_per_point))
        folded_flux_odd_har_array = np.empty((len(target_indexes), self.input_sizes[3], self.measurements_per_point))
        global_flux_array_err = np.empty((len(target_indexes), self.input_sizes[2], self.measurements_per_point))
        folded_flux_even_array_err = np.empty((len(target_indexes), self.input_sizes[3], self.measurements_per_point))
        folded_flux_odd_array_err = np.empty((len(target_indexes), self.input_sizes[3], self.measurements_per_point))
        folded_flux_even_subhar_array_err = np.empty((len(target_indexes), self.input_sizes[3], self.measurements_per_point))
        folded_flux_odd_subhar_array_err = np.empty((len(target_indexes), self.input_sizes[3], self.measurements_per_point))
        folded_flux_even_har_array_err = np.empty((len(target_indexes), self.input_sizes[3], self.measurements_per_point))
        folded_flux_odd_har_array_err = np.empty((len(target_indexes), self.input_sizes[3], self.measurements_per_point))
        folded_centroids_array = np.empty((len(target_indexes), self.input_sizes[3], 2))
        folded_centroids_array_err = np.empty((len(target_indexes), self.input_sizes[3], 2))
        folded_og_array = np.empty((len(target_indexes), self.input_sizes[3], 1))
        folded_og_array_err = np.empty((len(target_indexes), self.input_sizes[3], 1))
        batch_data_values = np.empty((len(target_indexes), 1))
        i = 0
        for df_index, target_row in injected_objects_df.iterrows():
            if self.fixed_target_id is not None:
                target_row = self.injected_objects_df[self.injected_objects_df['object_id'] == self.fixed_target_id[0]]
                target_row = target_row[target_row['period'] == self.fixed_target_id[1]]
            #TODO refactor to use mission for object_id
            object_id = target_row['object_id'].split(' ')
            mission_id = object_id[0]
            target_id = int(object_id[1].split('_')[0])
            period = target_row['period']
            lcs_dir = self.kics_lcs_dir if mission_id == 'KIC' else self.ete6_lcs_dir if 'OI' not in self.injected_objects_df else self.tess_lcs_dir
            if mission_id == 'KIC' or 'OI' in self.injected_objects_df:
                file_prefix = lcs_dir + '/' + mission_id + '_' + str(target_id) + '_' + str(round(period, 2))
            else:
                file_prefix = lcs_dir + '/' + mission_id + ' ' + str(target_id) + '_' + str(round(period, 2))
            if 'pl_orbper' in target_row and not np.isnan(target_row['pl_orbper']):
                period = target_row['pl_orbper']
            elif 'koi_period' in target_row and not np.isnan(target_row['koi_period']):
                period = target_row['koi_period']
            elif 'tce_period_x' in target_row and not np.isnan(target_row['tce_period_x']):
                period = target_row['tce_period_x']
            elif 'tce_period_y' in target_row and not np.isnan(target_row['tce_period_y']):
                period = target_row['tce_period_y']
            #period = target_row['koi_period']
            epoch = target_row['epoch']
            if 'pl_tranmid' in target_row and not np.isnan(target_row['pl_tranmid']):
                epoch = target_row['pl_tranmid'] - 2454833.0
            elif 'koi_time0bk' in target_row and not np.isnan(target_row['koi_time0bk']):
                epoch = target_row['koi_time0bk']
            elif 'tce_time0bk' in target_row and not np.isnan(target_row['tce_time0bk']):
                epoch = target_row['tce_time0bk']
            #epoch = target_row['koi_time0bk']
            duration = target_row['duration(h)'] / 24
            metrics_short_filename = file_prefix + '_metrics_short.csv'
            metrics_long_filename = file_prefix + '_metrics_long.csv'
            if 'tce_ptemp_stat' in target_row and not np.isnan(target_row['tce_ptemp_stat']):
                temperature_stat = target_row['tce_ptemp_stat']
            elif mission_id == 'TIC':
                temperature_stat = 0
                if os.path.exists(metrics_short_filename):
                    metrics_df = pd.read_csv(metrics_short_filename)
                    temperature_stat = metrics_df.loc[metrics_df['name'] == 'temp_stat'].iloc[0]['value']
                if os.path.exists(metrics_long_filename):
                    metrics_df = pd.read_csv(metrics_long_filename)
                    if temperature_stat is None:
                        temperature_stat = metrics_df.loc[metrics_df['name'] == 'temp_stat'].iloc[0]['value']
                    else:
                        temperature_stat = np.nanmean([temperature_stat,
                                                    metrics_df.loc[metrics_df['name'] == 'temp_stat'].iloc[0][
                                                        'value']])
            if 'tce_albedo_stat' in target_row and not np.isnan(target_row['tce_albedo']):
                albedo_stat = target_row['tce_albedo_stat']
            elif mission_id == 'TIC':
                albedo_stat = 0
                if os.path.exists(metrics_short_filename):
                    metrics_df = pd.read_csv(metrics_short_filename)
                    albedo_stat = metrics_df.loc[metrics_df['name'] == 'albedo_stat'].iloc[0]['value']
                if os.path.exists(metrics_long_filename):
                    metrics_df = pd.read_csv(metrics_long_filename)
                    if albedo_stat is None:
                        albedo_stat = metrics_df.loc[metrics_df['name'] == 'albedo_stat'].iloc[0]['value']
                    else:
                        albedo_stat = np.nanmean([albedo_stat,
                                                    metrics_df.loc[metrics_df['name'] == 'albedo_stat'].iloc[0][
                                                        'value']])
            if 'boot_fap' in target_row and not np.isnan(target_row['boot_fap']):
                bootstrap_fap = target_row['boot_fap']
                if np.isnan(bootstrap_fap) or bootstrap_fap <= 0:
                    bootstrap_fap = self.zero_epsilon
            elif mission_id == 'TIC':
                bootstrap_fap = 1 - self.zero_epsilon
                if os.path.exists(metrics_short_filename):
                    metrics_df = pd.read_csv(metrics_short_filename)
                    bootstrap_fap = metrics_df.loc[metrics_df['name'] == 'bootstrap_fap'].iloc[0]['value']
                if os.path.exists(metrics_long_filename):
                    metrics_df = pd.read_csv(metrics_long_filename)
                    if bootstrap_fap is None:
                        bootstrap_fap = metrics_df.loc[metrics_df['name'] == 'bootstrap_fap'].iloc[0]['value']
                    else:
                        bootstrap_fap = np.nanmean([bootstrap_fap,
                                                    metrics_df.loc[metrics_df['name'] == 'bootstrap_fap'].iloc[0][
                                                        'value']])
            duration_to_period = duration / period
            type = target_row['type']
            batch_data_values[i] = [float(x) for x in self.type_to_label[type]]
            if self.from_arrays:
                try:
                    scalar_values[i] = [[e] for e in np.loadtxt(file_prefix + '_input_scalar_values.csv', delimiter=',')]
                except Exception as e:
                    logging.info("FAILED FILE PREFIX: %s", file_prefix)
                    raise e
                star_array[i] = [[e] for e in np.loadtxt(file_prefix + '_input_star.csv', delimiter=',')]
                global_flux_array[i] = np.loadtxt(file_prefix + '_input_global.csv', delimiter=',').reshape(-1, 1)
                global_flux_array_err[i] = np.loadtxt(file_prefix + '_input_global_err.csv', delimiter=',').reshape(-1, 1)
                folded_flux_even_array[i] = np.loadtxt(file_prefix + '_input_even.csv', delimiter=',').reshape(-1, 1)
                folded_flux_even_array_err[i] = np.loadtxt(file_prefix + '_input_even_err.csv', delimiter=',').reshape(-1, 1)
                folded_flux_odd_array[i] = np.loadtxt(file_prefix + '_input_odd.csv', delimiter=',').reshape(-1, 1)
                folded_flux_odd_array_err[i] = np.loadtxt(file_prefix + '_input_odd_err.csv', delimiter=',').reshape(-1, 1)
                folded_flux_even_subhar_array[i] = np.loadtxt(file_prefix + '_input_even_sh.csv', delimiter=',').reshape(-1, 1)
                folded_flux_even_subhar_array_err[i] = np.loadtxt(file_prefix + '_input_even_sh_err.csv', delimiter=',').reshape(-1, 1)
                folded_flux_odd_subhar_array[i] = np.loadtxt(file_prefix + '_input_odd_sh.csv', delimiter=',').reshape(-1, 1)
                folded_flux_odd_subhar_array_err[i] = np.loadtxt(file_prefix + '_input_odd_sh_err.csv', delimiter=',').reshape(-1, 1)
                folded_flux_even_har_array[i] = np.loadtxt(file_prefix + '_input_even_h.csv', delimiter=',').reshape(-1, 1)
                folded_flux_even_har_array_err[i] = np.loadtxt(file_prefix + '_input_even_h_err.csv', delimiter=',').reshape(-1, 1)
                folded_flux_odd_har_array[i] = np.loadtxt(file_prefix + '_input_odd_h.csv', delimiter=',').reshape(-1, 1)
                folded_flux_odd_har_array_err[i] = np.loadtxt(file_prefix + '_input_odd_h_err.csv', delimiter=',').reshape(-1, 1)
                folded_centroids_array[i] = np.loadtxt(file_prefix + '_input_centroids.csv', delimiter=',')
                folded_centroids_array_err[i] = np.loadtxt(file_prefix + '_input_centroids_err.csv', delimiter=',')
                folded_og_array[i] = np.loadtxt(file_prefix + '_input_og.csv', delimiter=',').reshape(-1, 1)
                folded_og_array_err[i] = np.loadtxt(file_prefix + '_input_og_err.csv', delimiter=',').reshape(-1, 1)
            else:
                #TODO refactor to use mission for _lc files
                lc_filename = file_prefix + '_lc.csv'
                centroids_filename = file_prefix + '_cent.csv'
                og_filename = file_prefix + '_og.csv'
                lc_short_filename = file_prefix + '_lc_short.csv'
                centroids_short_filename = file_prefix + '_cent_short.csv'
                og_short_filename = file_prefix + '_og_short.csv'
                lc_long_filename = file_prefix + '_lc_long.csv'
                centroids_long_filename = file_prefix + '_cent_long.csv'
                og_long_filename = file_prefix + '_og_long.csv'
                lc_df = None
                centroids_df = None
                og_df = None
                if os.path.exists(lc_long_filename):
                    read_df = pd.read_csv(lc_long_filename, usecols=['#time', 'flux_0'], low_memory=True)
                    lc_df = read_df if lc_df is None else lc_df.append(read_df, ignore_index=True)
                if os.path.exists(lc_short_filename):
                    read_df = pd.read_csv(lc_short_filename, usecols=['#time', 'flux_0'], low_memory=True)
                    lc_df = read_df if lc_df is None else lc_df.append(read_df, ignore_index=True)
                if lc_df is None:
                    logging.warning("No curve for target " + file_prefix)
                    raise ValueError("No curve for target " + file_prefix)
                if os.path.exists(centroids_long_filename):
                    read_centroids_df = pd.read_csv(centroids_long_filename, usecols=['time', 'centroids_ra', 'centroids_dec'],
                                                    low_memory=True)
                    read_centroids_df = read_centroids_df.sort_values(by=['time'])
                    if len(read_centroids_df) > 0:
                        read_centroids_df['centroids_ra'], _ = LcbuilderHelper.detrend(read_centroids_df['time'].to_numpy(),
                                                                read_centroids_df['centroids_ra'].to_numpy(), duration * 4,
                                                                check_cadence=True)
                        read_centroids_df['centroids_dec'], _ = LcbuilderHelper.detrend(read_centroids_df['time'].to_numpy(),
                                                                read_centroids_df['centroids_dec'].to_numpy(), duration * 4,
                                                                check_cadence=True)
                        centroids_df = read_centroids_df if centroids_df is None \
                            else centroids_df.append(read_centroids_df, ignore_index=True)
                if os.path.exists(og_long_filename):
                    read_og_df = pd.read_csv(og_long_filename, usecols=['time', 'og_flux', 'halo_flux', 'core_flux'],
                                             low_memory=True)
                    read_og_df = self.compute_og_df(read_og_df, object_id, duration)
                    og_df = read_og_df if og_df is None else og_df.append(read_og_df, ignore_index=True)
                if os.path.exists(centroids_short_filename):
                    read_centroids_df = pd.read_csv(centroids_short_filename, usecols=['time', 'centroids_ra', 'centroids_dec'],
                                                    low_memory=True)
                    if len(read_centroids_df) > 0:
                        read_centroids_df = read_centroids_df.sort_values(by=['time'])
                        read_centroids_df['centroids_ra'], _ = LcbuilderHelper.detrend(read_centroids_df['time'].to_numpy(),
                                                                read_centroids_df['centroids_ra'].to_numpy(), duration * 4,
                                                                check_cadence=True)
                        read_centroids_df['centroids_dec'], _ = LcbuilderHelper.detrend(read_centroids_df['time'].to_numpy(),
                                                                read_centroids_df['centroids_dec'].to_numpy(), duration * 4,
                                                                check_cadence=True)
                    centroids_df = read_centroids_df if centroids_df is None \
                        else centroids_df.append(read_centroids_df, ignore_index=True)
                if os.path.exists(og_short_filename):
                    read_og_df = pd.read_csv(og_short_filename, usecols=['time', 'og_flux', 'halo_flux', 'core_flux'],
                                             low_memory=True)
                    read_og_df = self.compute_og_df(read_og_df, object_id, duration)
                    og_df = read_og_df if og_df is None else og_df.append(read_og_df, ignore_index=True)
                lc_df = lc_df.sort_values(by=['#time'])
                if self.mask_previous_signals:
                    lc_df = self.mask_other_signals(lc_df, target_row, '#time')
                    centroids_df = self.mask_other_signals(centroids_df, target_row, 'time')
                    og_df = self.mask_other_signals(og_df, target_row, 'time')
                centroids_df['time'] = self.fold(centroids_df['time'].to_numpy(), period, epoch + period / 2)
                og_df['time'] = self.fold(og_df['time'].to_numpy(), period, epoch + period / 2)
                centroids_df = centroids_df.sort_values(by=['time'])
                og_df = og_df.sort_values(by=['time'])
                centroids_df = self._prepare_input_centroids(centroids_df)
                og_df = self._prepare_input_og(og_df)
                og_df = og_df[(og_df['time'] > 0.5 - duration_to_period * 3) & (
                        og_df['time'] < 0.5 + duration_to_period * 3)]
                centroids_df = centroids_df[(centroids_df['time'] > 0.5 - duration_to_period * 3) & (
                        centroids_df['time'] < 0.5 + duration_to_period * 3)]
                folded_centroids_array[i], folded_centroids_array_err[i] = self.bin_by_time(centroids_df.to_numpy(), self.input_sizes[3], target_row['object_id'])
                folded_og_array[i], folded_og_array_err[i] = self.bin_by_time(og_df.to_numpy(), self.input_sizes[3], target_row['object_id'])
                #TODO refactor to use mission for star files
                star_filename = lcs_dir + '/' + mission_id + ' ' + str(target_id) + '_star.csv'
                star_df = pd.read_csv(star_filename, index_col=False)
                star_neighbours_df = pd.read_csv(star_filename,
                                                 usecols=['Teff', 'lum', 'v', 'j', 'k', 'h', 'radius', 'mass',
                                                          'dist_arcsec'], index_col=False)
                star_neighbours_df = self._prepare_input_neighbour_stars(star_neighbours_df)
                star_df, ra, dec = self._prepare_input_star(target_row, star_df)
                lc_df, good_transits_count, transits_count = self._prepare_input_lc(lc_df, period, epoch, duration)
                not_null_times_args = np.argwhere(lc_df['#time'].to_numpy() > 0).flatten()
                lc_df = lc_df.iloc[not_null_times_args]
                offset_long_filename = file_prefix + '_offset_long.csv'
                offset_short_filename = file_prefix + '_offset_short.csv'
                offset_short_df = None
                offset_long_df = None
                offset_ra = None
                offset_dec = None
                offset_ra_err = None
                offset_dec_err = None
                if os.path.exists(offset_long_filename):
                    offset_long_df = pd.read_csv(offset_long_filename, low_memory=True)
                    row = offset_long_df[offset_long_df['name'] == self.offset_mode].iloc[0]
                    offset_ra = row['ra']
                    offset_dec = row['dec']
                    offset_ra_err = row['ra_err']
                    offset_dec_err = row['dec_err']
                if os.path.exists(offset_short_filename):
                    offset_short_df = pd.read_csv(offset_short_filename, low_memory=True)
                    row = offset_short_df[offset_short_df['name'] == self.offset_mode].iloc[0]
                    row_mean = offset_short_df[offset_short_df['name'] == 'mean'].iloc[0]
                    offset_ra = row['ra'] if offset_long_df is None or np.isnan(offset_ra) else np.mean(
                        [row['ra'], offset_ra])
                    offset_dec = row['dec'] if offset_long_df is None or np.isnan(offset_dec) else np.mean(
                        [row['dec'], offset_dec])
                    offset_ra_err = self.zero_epsilon if np.isnan(offset_ra) else offset_ra_err
                    offset_dec_err = self.zero_epsilon if np.isnan(offset_dec) else offset_dec_err
                    row_ra_err = row['ra_err'] if not np.isnan(row['ra_err']) and row['ra_err'] > 0 else row_mean['ra_err']
                    row_dec_err = row['dec_err'] if not np.isnan(row['dec_err']) and row['dec_err'] > 0 else row_mean['dec_err']
                    offset_ra_err = row_ra_err if offset_long_df is None or np.isnan(offset_ra_err) else np.sqrt(
                        offset_ra_err ** 2 + row_ra_err ** 2)
                    offset_dec_err = row_dec_err if offset_long_df is None or np.isnan(offset_dec_err) else np.sqrt(
                        offset_dec_err ** 2 + row_dec_err ** 2)
                target_dist = np.sqrt((offset_ra - ra) ** 2 + (offset_dec - dec) ** 2)
                if target_dist < self.zero_epsilon:
                    logging.info(f"target_dist was zero for {file_prefix}")
                    target_dist = random.uniform(self.zero_epsilon, 1 / 3600 * LcbuilderHelper.mission_pixel_size(MISSION_TESS if mission_id == MISSION_ID_TESS else MISSION_KEPLER))
                target_dist = target_dist if target_dist > 0 else self.zero_epsilon
                offset_err = offset_ra_err if offset_ra_err > offset_dec_err else offset_dec_err
                offset_err = offset_err if offset_err > 0 else target_dist * 2
                offset_err = offset_err / (target_dist * 2)
                offset_err = 1 - self.zero_epsilon if offset_err >= 1 else offset_err
                offset_err = offset_err if offset_err > 0 else self.zero_epsilon
                good_transits_count_norm = good_transits_count / 20
                good_transits_count_norm = good_transits_count_norm if good_transits_count_norm < 1 else 1 - self.zero_epsilon
                good_transits_ratio = good_transits_count / transits_count if transits_count > 0 else self.zero_epsilon
                good_transits_ratio = good_transits_ratio if good_transits_ratio < 1 else 1 - self.zero_epsilon
                depth = target_row['depth_primary'] / 1e6
                planet_radius = target_row['radius(earth)']
                # if not np.isnan(star_df['radius']) and not np.isnan(target_row['depth_primary']):
                #     planet_radius = np.sqrt(depth * LcbuilderHelper.convert_from_to(star_df['radius'], u.R_sun, u.R_earth) ** 2)
                planet_radius = planet_radius / 300
                depth = depth if depth < 1 else 1 - self.zero_epsilon
                planet_radius = planet_radius if planet_radius < 1 else 1 - self.zero_epsilon
                #['ld_a', 'ld_b', 'Teff', 'lum', 'logg', 'radius', 'mass', 'v', 'j', 'h', 'k'])
                neighbours_array = star_neighbours_df.to_numpy().flatten()
                star_array[i] = np.transpose([star_df.to_numpy()])
                time = lc_df['#time'].to_numpy()
                # Global flux
                # Shifting data 1/4 so that main transit and possible occultation don't get cut by the borders
                lc_df_sorted_fold = lc_df.copy()
                lc_df_sorted_fold['#time'] = self.fold(time, period, epoch + period / 4)
                lc_df_sorted_fold = lc_df_sorted_fold.sort_values(by=['#time'])
                global_flux_array[i], global_flux_array_err[i] = \
                    self.bin_by_time(lc_df_sorted_fold.to_numpy(), self.input_sizes[2], target_row['object_id'])
                # Focus flux even (secondary event)
                lc_df_focus = lc_df.copy()
                lc_df_focus['#time'] = self.fold(time, period, epoch)
                lc_df_focus = lc_df_focus.sort_values(by=['#time'])
                lc_df_focus = lc_df_focus[(lc_df_focus['#time'] > 0.5 - duration_to_period * 3) &
                                          (lc_df_focus['#time'] < 0.5 + duration_to_period * 3)]
                folded_flux_even_array[i], folded_flux_even_array_err[i] = \
                    self.bin_by_time(lc_df_focus.to_numpy(), self.input_sizes[3], target_row['object_id'])
                # Focus flux odd
                lc_df_focus = lc_df.copy()
                lc_df_focus['#time'] = self.fold(time, period, epoch + period / 2)
                lc_df_focus = lc_df_focus.sort_values(by=['#time'])
                lc_df_focus = lc_df_focus[(lc_df_focus['#time'] > 0.5 - duration_to_period * 3) &
                                          (lc_df_focus['#time'] < 0.5 + duration_to_period * 3)]
                folded_flux_odd_array[i], folded_flux_odd_array_err[i] = \
                    self.bin_by_time(lc_df_focus.to_numpy(), self.input_sizes[4], target_row['object_id'])
                # Focus flux harmonic even
                lc_df_focus = lc_df.copy()
                lc_df_focus['#time'] = self.fold(time, period * 2, epoch)
                lc_df_focus = lc_df_focus.sort_values(by=['#time'])
                lc_df_focus = lc_df_focus[(lc_df_focus['#time'] > 0.5 - duration_to_period * 1.5) &
                                          (lc_df_focus['#time'] < 0.5 + duration_to_period * 1.5)]
                oot_lc_df_focus = lc_df_focus[((lc_df_focus['#time'] <= 0.5 - duration_to_period / 4) |
                                           (lc_df_focus['#time'] >= 0.5 + duration_to_period / 4))]
                it_lc_df_focus = lc_df_focus[((lc_df_focus['#time'] > 0.5 - duration_to_period / 4) &
                                           (lc_df_focus['#time'] < 0.5 + duration_to_period / 4))]
                depth_even = (oot_lc_df_focus['flux_0'].mean() - it_lc_df_focus['flux_0'].mean()) * np.sqrt(len(it_lc_df_focus)) / oot_lc_df_focus['flux_0'].std()
                folded_flux_even_har_array[i], folded_flux_even_har_array_err[i] = \
                    self.bin_by_time(lc_df_focus.to_numpy(), self.input_sizes[7], target_row['object_id'])
                # Focus flux harmonic odd
                lc_df_focus = lc_df.copy()
                lc_df_focus['#time'] = self.fold(time, period * 2, epoch + period)
                lc_df_focus = lc_df_focus.sort_values(by=['#time'])
                lc_df_focus = lc_df_focus[(lc_df_focus['#time'] > 0.5 - duration_to_period * 1.5) &
                                          (lc_df_focus['#time'] < 0.5 + duration_to_period * 1.5)]
                oot_lc_df_focus = lc_df_focus[((lc_df_focus['#time'] <= 0.5 - duration_to_period / 4) |
                                               (lc_df_focus['#time'] >= 0.5 + duration_to_period / 4))]
                it_lc_df_focus = lc_df_focus[((lc_df_focus['#time'] > 0.5 - duration_to_period / 4) &
                                              (lc_df_focus['#time'] < 0.5 + duration_to_period / 4))]
                depth_odd = (oot_lc_df_focus['flux_0'].mean() - it_lc_df_focus['flux_0'].mean()) * np.sqrt(
                    len(it_lc_df_focus)) / oot_lc_df_focus['flux_0'].std()
                folded_flux_odd_har_array[i], folded_flux_odd_har_array_err[i] = \
                    self.bin_by_time(lc_df_focus.to_numpy(), self.input_sizes[8], target_row['object_id'])
                # Focus flux sub-harmonic even
                lc_df_focus = pd.DataFrame(columns=['#time', 'flux_0'])
                time, flux0, _ = LcbuilderHelper.mask_transits(time,
                                                                lc_df.copy()['flux_0'].to_numpy(), period, duration * 6,
                                                                epoch)
                lc_df_focus['#time'] = time
                lc_df_focus['flux_0'] = flux0
                lc_df_focus['#time'] = self.fold(time, period / 2, epoch)
                lc_df_focus = lc_df_focus.sort_values(by=['#time'])
                lc_df_focus = lc_df_focus[(lc_df_focus['#time'] > 0.5 - duration_to_period * 3) &
                                          (lc_df_focus['#time'] < 0.5 + duration_to_period * 3)]
                folded_flux_even_subhar_array[i], folded_flux_even_subhar_array_err[i] = \
                    self.bin_by_time(lc_df_focus.to_numpy(), self.input_sizes[5], target_row['object_id'])
                # Focus flux sub-harmonic odd
                lc_df_focus = pd.DataFrame(columns=['#time', 'flux_0'])
                lc_df_focus['#time'] = time
                lc_df_focus['flux_0'] = flux0
                lc_df_focus['#time'] = self.fold(time, period / 2, epoch + period / 4)
                lc_df_focus = lc_df_focus.sort_values(by=['#time'])
                lc_df_focus = lc_df_focus[(lc_df_focus['#time'] > 0.5 - duration_to_period * 3) &
                                          (lc_df_focus['#time'] < 0.5 + duration_to_period * 3)]
                folded_flux_odd_subhar_array[i], folded_flux_odd_subhar_array_err[i] = \
                    self.bin_by_time(lc_df_focus.to_numpy(), self.input_sizes[6], target_row['object_id'])
                oe_factor = np.abs(depth_even - depth_odd)
                oe_factor = oe_factor / 5 if oe_factor < 5 else 1 - self.zero_epsilon
                oe_factor = oe_factor if not np.isnan(oe_factor) and oe_factor > 0 else self.zero_epsilon
                if np.isnan(temperature_stat):
                    temperature_stat = 0
                elif temperature_stat < -10:
                    temperature_stat = -10
                elif temperature_stat > 10:
                    temperature_stat = 10
                temperature_stat = (temperature_stat + 10) / 20
                temperature_stat = temperature_stat if temperature_stat > 0 else self.zero_epsilon
                temperature_stat = temperature_stat if temperature_stat < 1 else 1 - self.zero_epsilon
                if np.isnan(albedo_stat):
                    albedo_stat = 0
                elif albedo_stat < -10:
                    albedo_stat = -10
                elif albedo_stat > 10:
                    albedo_stat = 10
                albedo_stat = (albedo_stat + 10) / 20
                albedo_stat = albedo_stat if albedo_stat > 0 else self.zero_epsilon
                albedo_stat = albedo_stat if albedo_stat < 1 else 1 - self.zero_epsilon
                scalar_values[i] = np.transpose([[period / 1200 if period < 1200 else 1,
                                                  duration / 15,
                                                  depth,
                                                  planet_radius,
                                                  good_transits_count_norm,
                                                  good_transits_ratio,
                                                  target_dist if not np.isnan(target_dist) else self.zero_epsilon,
                                                  offset_err,
                                                  bootstrap_fap,
                                                  oe_factor,
                                                  temperature_stat,
                                                  albedo_stat]])
            self.assert_in_range(object_id, scalar_values[i], None)
            self.assert_in_range(object_id, global_flux_array[i], global_flux_array_err[i])
            self.assert_in_range(object_id, star_array[i], None)
            self.assert_in_range(object_id, folded_flux_even_array[i], folded_flux_even_array_err[i])
            self.assert_in_range(object_id, folded_flux_odd_array[i], folded_flux_odd_array_err[i])
            self.assert_in_range(object_id, folded_flux_even_subhar_array[i], folded_flux_even_subhar_array_err[i])
            self.assert_in_range(object_id, folded_flux_odd_subhar_array[i], folded_flux_odd_subhar_array_err[i])
            self.assert_in_range(object_id, folded_flux_even_har_array[i], folded_flux_even_har_array_err[i])
            self.assert_in_range(object_id, folded_flux_odd_har_array[i], folded_flux_odd_har_array_err[i])
            self.assert_in_range(object_id, folded_centroids_array[i], folded_centroids_array_err[i])
            self.assert_in_range(object_id, folded_og_array[i], folded_og_array_err[i])
                # logging.info("GENERATOR Star Inputs max " + str(np.max(star_array[i])) + " and min " +
                #              str(np.min(star_array[i])))
                # logging.info("GENERATOR Global Flux Inputs max " + str(np.max(global_flux_array[i])) + " and min " +
                #              str(np.min(global_flux_array[i])))
                # logging.info("GENERATOR Folded Even Inputs max " + str(np.max(folded_flux_even_array[i])) + " and min " +
                #              str(np.min(folded_flux_even_array[i])))
                # logging.info("GENERATOR Folded Odd Inputs max " + str(np.max(folded_flux_odd_array[i])) + " and min " +
                #              str(np.min(folded_flux_odd_array[i])))
                # logging.info("GENERATOR Folded Even Subhar Inputs max " + str(np.max(folded_flux_even_subhar_array[i])) +
                #              " and min " + str(np.min(folded_flux_even_subhar_array[i])))
                # logging.info("GENERATOR Folded Odd Subhar Inputs max " + str(np.max(folded_flux_odd_subhar_array[i])) +
                #              " and min " + str(np.min(folded_flux_odd_subhar_array[i])))
                # logging.info("GENERATOR Folded Even Har Inputs max " + str(np.max(folded_flux_even_har_array[i])) +
                #              " and min " + str(np.min(folded_flux_even_har_array[i])))
                # logging.info("GENERATOR Folded Odd Har Inputs max " + str(np.max(folded_flux_odd_har_array[i])) +
                #              " and min " + str(np.min(folded_flux_odd_har_array[i])))
            self._plot_input(np.transpose(global_flux_array[i])[0], target_row['object_id'] + "_" + type, "global")
            self._plot_input(np.transpose(folded_flux_even_array[i])[0], target_row['object_id'] + "_" + type, "even")
            self._plot_input(np.transpose(folded_flux_odd_array[i])[0], target_row['object_id'] + "_" + type, "odd")
            self._plot_input(np.transpose(folded_flux_even_har_array[i])[0], target_row['object_id'] + "_" + type, "even_har")
            self._plot_input(np.transpose(folded_flux_odd_har_array[i])[0], target_row['object_id'] + "_" + type, "odd_har")
            self._plot_input(np.transpose(folded_og_array[i])[0], target_row['object_id'] + "_" + type, "OG")
            self._plot_input(np.transpose(folded_centroids_array[i])[0], target_row['object_id'] + "_" + type, "CENTROIDS RA")
            self._plot_input(np.transpose(folded_centroids_array[i])[1], target_row['object_id'] + "_" + type, "CENTROIDS DEC")
            self._plot_numeric_values(object_id, scalar_values[i], star_array[i])
            if self.store_arrays:
                logging.info("Storing arrays into prefix " + file_prefix)
                np.savetxt(file_prefix + '_input_scalar_values.csv', scalar_values[i], delimiter=',')
                np.savetxt(file_prefix + '_input_star.csv', star_array[i], delimiter=',')
                np.savetxt(file_prefix + '_input_global.csv', global_flux_array[i], delimiter=',')
                np.savetxt(file_prefix + '_input_global_err.csv', global_flux_array_err[i], delimiter=',')
                np.savetxt(file_prefix + '_input_even.csv', folded_flux_even_array[i], delimiter=',')
                np.savetxt(file_prefix + '_input_even_err.csv', folded_flux_even_array_err[i], delimiter=',')
                np.savetxt(file_prefix + '_input_odd.csv', folded_flux_odd_array[i], delimiter=',')
                np.savetxt(file_prefix + '_input_odd_err.csv', folded_flux_odd_array_err[i], delimiter=',')
                np.savetxt(file_prefix + '_input_even_sh.csv', folded_flux_even_subhar_array[i], delimiter=',')
                np.savetxt(file_prefix + '_input_even_sh_err.csv', folded_flux_even_subhar_array_err[i], delimiter=',')
                np.savetxt(file_prefix + '_input_odd_sh.csv', folded_flux_odd_subhar_array[i], delimiter=',')
                np.savetxt(file_prefix + '_input_odd_sh_err.csv', folded_flux_odd_subhar_array_err[i], delimiter=',')
                np.savetxt(file_prefix + '_input_even_h.csv', folded_flux_even_har_array[i], delimiter=',')
                np.savetxt(file_prefix + '_input_even_h_err.csv', folded_flux_even_har_array_err[i], delimiter=',')
                np.savetxt(file_prefix + '_input_odd_h.csv', folded_flux_odd_har_array[i], delimiter=',')
                np.savetxt(file_prefix + '_input_odd_h_err.csv', folded_flux_odd_har_array_err[i], delimiter=',')
                np.savetxt(file_prefix + '_input_centroids.csv', folded_centroids_array[i], delimiter=',')
                np.savetxt(file_prefix + '_input_centroids_err.csv', folded_centroids_array_err[i], delimiter=',')
                np.savetxt(file_prefix + '_input_og.csv', folded_og_array[i], delimiter=',')
                np.savetxt(file_prefix + '_input_og_err.csv', folded_og_array_err[i], delimiter=',')
            if self.explain:
                if 'EXPLAIN_PERIOD_VALUE' in target_row and not np.isnan(target_row['EXPLAIN_PERIOD_VALUE']):
                    scalar_values[i, 0] = target_row['EXPLAIN_PERIOD_VALUE']
                if 'EXPLAIN_DURATION_VALUE' in target_row and not np.isnan(target_row['EXPLAIN_DURATION_VALUE']):
                    scalar_values[i, 1] = target_row['EXPLAIN_DURATION_VALUE']
                if 'EXPLAIN_RADIUS_VALUE' in target_row and not np.isnan(target_row['EXPLAIN_RADIUS_VALUE']):
                    scalar_values[i, 3] = target_row['EXPLAIN_RADIUS_VALUE']
                if 'EXPLAIN_GTCN_VALUE' in target_row and not np.isnan(target_row['EXPLAIN_GTCN_VALUE']):
                    scalar_values[i, 4] = target_row['EXPLAIN_GTCN_VALUE']
                if 'EXPLAIN_GTR_VALUE' in target_row and not np.isnan(target_row['EXPLAIN_GTR_VALUE']):
                    scalar_values[i, 5] = target_row['EXPLAIN_GTR_VALUE']
                if 'EXPLAIN_OFFSET_VALUE' in target_row and not np.isnan(target_row['EXPLAIN_OFFSET_VALUE']):
                    scalar_values[i, 7] = target_row['EXPLAIN_OFFSET_VALUE']
                if 'EXPLAIN_ODDEVEN_VALUE' in target_row and not np.isnan(target_row['EXPLAIN_ODDEVEN_VALUE']):
                    scalar_values[i, 8] = target_row['EXPLAIN_ODDEVEN_VALUE']
                if 'EXPLAIN_BFAP_VALUE' in target_row and not np.isnan(target_row['EXPLAIN_BFAP_VALUE']):
                    scalar_values[i, 9] = target_row['EXPLAIN_BFAP_VALUE']
                if 'EXPLAIN_TEMP_VALUE' in target_row and not np.isnan(target_row['EXPLAIN_TEMP_VALUE']):
                    scalar_values[i, 10] = target_row['EXPLAIN_TEMP_VALUE']
                if 'EXPLAIN_ALBEDO_VALUE' in target_row and not np.isnan(target_row['EXPLAIN_ALBEDO_VALUE']):
                    scalar_values[i, 11] = target_row['EXPLAIN_ALBEDO_VALUE']
                if 'EXPLAIN_STARTEFF_VALUE' in target_row and not np.isnan(target_row['EXPLAIN_STARTEFF_VALUE']):
                    star_array[i, 2] = target_row['EXPLAIN_STARTEFF_VALUE']
                if 'EXPLAIN_STARRAD_VALUE' in target_row and not np.isnan(target_row['EXPLAIN_STARRAD_VALUE']):
                    star_array[i, 6] = target_row['EXPLAIN_STARRAD_VALUE']
                if 'EXPLAIN_GLOBAL_VIEW' in target_row and not np.isnan(target_row['EXPLAIN_GLOBAL_VIEW']):
                    global_flux_array[i] = np.zeros(global_flux_array[i].shape)
                if 'EXPLAIN_MAIN_VIEW' in target_row and not np.isnan(target_row['EXPLAIN_MAIN_VIEW']):
                    folded_flux_odd_array[i] = np.zeros(folded_flux_odd_array[i].shape)
                if 'EXPLAIN_SECONDARY_VIEW' in target_row and not np.isnan(target_row['EXPLAIN_SECONDARY_VIEW']):
                    folded_flux_even_array[i] = np.zeros(folded_flux_even_array[i].shape)
                if 'EXPLAIN_ODD_VIEW' in target_row and not np.isnan(target_row['EXPLAIN_ODD_VIEW']):
                    folded_flux_odd_har_array[i] = np.zeros(folded_flux_odd_har_array[i].shape)
                if 'EXPLAIN_EVEN_VIEW' in target_row and not np.isnan(target_row['EXPLAIN_EVEN_VIEW']):
                    folded_flux_even_har_array[i] = np.zeros(folded_flux_even_har_array[i].shape)
                if 'EXPLAIN_CENTROIDS_RA_VIEW' in target_row and not np.isnan(target_row['EXPLAIN_CENTROIDS_RA_VIEW']):
                    folded_centroids_array[i, :, 0] = np.zeros(folded_centroids_array[i].shape[0])
                if 'EXPLAIN_CENTROIDS_DEC_VIEW' in target_row and not np.isnan(target_row['EXPLAIN_CENTROIDS_DEC_VIEW']):
                    folded_centroids_array[i, :, 1] = np.zeros(folded_centroids_array[i].shape[0])
                if 'EXPLAIN_OG_VIEW' in target_row and not np.isnan(target_row['EXPLAIN_OG_VIEW']):
                    folded_og_array[i] = np.zeros(folded_og_array[i].shape)
            i = i + 1
        return [scalar_values[:, [0, 1, 3, 4, 5, 8, 9, 10, 11]], scalar_values[:, [7]],
                star_array[:, [2, 6]], #Only Teff, Rad
                global_flux_array,
                folded_flux_even_array,
                folded_flux_odd_array,
                folded_flux_even_har_array,
                folded_flux_odd_har_array,
                folded_centroids_array,
                folded_og_array], \
            batch_data_values



import collections
import copy
import csv
import keras.backend as K
from keras.callbacks import CSVLogger
import numpy as np
from exoml.ml.callback.learning_rate import MultiOptimizer


class BatchAwareCsvLogger(CSVLogger):

    def __init__(self, filename, steps_per_epoch, separator=',', append=False):
        super().__init__(filename, separator, append)
        self.steps_per_epoch = steps_per_epoch
        self.current_epoch = 0

    def on_epoch_end(self, epoch, logs=None):
        self.__store_log(logs, self.steps_per_epoch - 1, self.current_epoch, False)

    def on_epoch_begin(self, epoch, logs=None):
        self.current_epoch = epoch

    def on_batch_end(self, batch, logs=None):
        self.__store_log(logs, batch, self.current_epoch)

    def __store_log(self, logs, batch, epoch, with_validation=True):
        formatted_logs = copy.deepcopy(logs) or {}

        def handle_value(k):
            is_zero_dim_ndarray = isinstance(k, np.ndarray) and k.ndim == 0
            if isinstance(k, str):
                return k
            elif isinstance(k, collections.abc.Iterable) and not is_zero_dim_ndarray:
                return '"[%s]"' % (', '.join(map(str, k)))
            else:
                return k
        keys = formatted_logs.keys()
        if self.model.stop_training:
            # We set NA so that csv parsers do not fail for this last epoch.
            formatted_logs = dict((k, formatted_logs[k]) if k in formatted_logs else (k, 'NA') for k in keys)
        validation_fields = ['val_' + key for key in keys] if with_validation else []
        metric_fieldnames = list(keys) + validation_fields
        if not self.writer:
            class CustomDialect(csv.excel):
                delimiter = self.sep
            lr_fieldnames = ['lr']
            if isinstance(self.model.optimizer, MultiOptimizer):
                lr_fieldnames = lr_fieldnames + ['lr_last']
            fieldnames = ['epoch', 'batch'] + lr_fieldnames +  metric_fieldnames
            self.writer = csv.DictWriter(
                self.csv_file,
                fieldnames=fieldnames,
                dialect=CustomDialect)
            if self.append_header:
                self.writer.writeheader()
        row_dict = {}
        if isinstance(self.model.optimizer, MultiOptimizer):
            learning_rate = K.eval(K.eval(self.model.optimizer.optimizer_specs[0]['optimizer'].lr))
            last_learning_rate = K.eval(K.eval(self.model.optimizer.optimizer_specs[-1]['optimizer'].lr))
            row_dict['lr_last'] = last_learning_rate
        else:
            learning_rate = K.eval(self.model.optimizer.lr)
        row_dict['epoch'] = epoch
        row_dict['batch'] = batch
        row_dict['lr'] = learning_rate
        for metric_key in metric_fieldnames:
            row_dict[metric_key] = handle_value(formatted_logs[metric_key][0]
                                                if isinstance(formatted_logs[metric_key], (list, np.ndarray))
                                                else formatted_logs[metric_key]) if metric_key in formatted_logs else np.nan
        self.writer.writerow(row_dict)
        self.csv_file.flush()



import dataclasses

import tensorflow as tf

from keras_core.src.optimizers.schedules.learning_rate_schedule import LearningRateSchedule
#from keras.optimizers.schedules import LearningRateSchedule


@dataclasses.dataclass(init=False)
class ExponentialRescaleDecay(LearningRateSchedule):
    def __init__(
            self,
            initial_learning_rate,
            decay_steps,
            decay_rate,
            restore_steps,
            restore_rate=1.5,
            staircase=False,
            name=None):
        """Applies exponential decay to the learning rate and a value recovery of the learning rate to prevent gradient
    vanishing.

    Args:
      initial_learning_rate: A scalar `float32` or `float64` `Tensor` or a
        Python number.  The initial learning rate.
      decay_steps: A scalar `int32` or `int64` `Tensor` or a Python number.
        Must be positive.  See the decay computation above.
      decay_rate: A scalar `float32` or `float64` `Tensor` or a
        Python number.  The decay rate.
      restore_steps: A scalar `float32` or `float64` `Tensor` or a
        Python number.  The number of steps to wait for a learning rate rescale
      restore_rate: A scalar `float32` or `float64` `Tensor` or a
        Python number.  The restore rate.
      staircase: Boolean.  If `True` decay the learning rate at discrete
        intervals
      name: String.  Optional name of the operation.  Defaults to
        'ExponentialDecay'.
    """
        super(ExponentialRescaleDecay, self).__init__()
        assert restore_steps > decay_steps
        self.initial_learning_rate = initial_learning_rate
        self.decay_steps = decay_steps
        self.decay_rate = decay_rate
        self.restore_steps = restore_steps
        self.restore_rate = restore_rate
        self.staircase = staircase
        self.name = name

    def __call__(self, step):
        with tf.name_scope(self.name or "ExponentialRescaleDecay") as name:
            initial_learning_rate = tf.convert_to_tensor(
                self.initial_learning_rate, dtype=tf.float64, name="initial_learning_rate")
            dtype = initial_learning_rate.dtype
            decay_steps = tf.cast(self.decay_steps, dtype)
            decay_rate = tf.cast(self.decay_rate, dtype)
            restore_rate = tf.cast(self.restore_rate, dtype)
            restore_steps = tf.cast(self.restore_steps, dtype)
            step_double = tf.cast(step, dtype)
            # Mod is enforcing float64
            steps_since_last_rescale = tf.math.mod(step_double, restore_steps)
            restores_done = tf.math.floordiv(step_double, restore_steps)
            one_tensor = tf.cast(tf.constant(1), dtype)
            restores_done = tf.cond(tf.math.equal(restores_done, 0),
                                    lambda: one_tensor,
                                    lambda: restores_done + one_tensor)
            restores_done_by_rate = tf.cond(tf.math.equal(restores_done, 1),
                                            lambda: restores_done,
                                            lambda: tf.divide(restores_done, restore_rate))
            p = steps_since_last_rescale / decay_steps
            if self.staircase:
                p = tf.floor(p)
            new_learning_rate = tf.divide(initial_learning_rate, restores_done_by_rate)
            new_learning_rate = tf.multiply(new_learning_rate, tf.pow(decay_rate, p), name=name)
            return new_learning_rate

    def get_config(self):
        return {
            "initial_learning_rate": self.initial_learning_rate,
            "decay_steps": self.decay_steps,
            "decay_rate": self.decay_rate,
            "restore_steps": self.restore_steps,
            "restore_rate": self.restore_rate,
            "staircase": self.staircase,
            "name": self.name
        }



import os
import math
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import sklearn.metrics as metrics
from typing import Dict, List, Tuple, Optional
from sklearn.utils import check_consistent_length, column_or_1d
from sklearn.calibration import calibration_curve


__all__ = [
    'compute_calibration_error',
    'create_binned_data',
    'get_bin_boundaries',
    'compute_binary_score',
    'compute_calibration_summary',
]

#File taken from https://github.com/ethen8181/machine-learning/blob/master/model_selection/prob_calibration/calibration_module/utils.py

def compute_calibration_error(
    y_true: np.ndarray,
    y_prob: np.ndarray,
    n_bins: int=15,
    round_digits: int=4) -> float:
    """
    Computes the calibration error for binary classification via binning
    data points into the specified number of bins. Samples with similar
    ``y_prob`` will be grouped into the same bin. The bin boundary is
    determined by having similar number of samples within each bin.

    Parameters
    ----------
    y_true : 1d ndarray
        Binary true targets.

    y_prob : 1d ndarray
        Raw probability/score of the positive class.

    n_bins : int, default 15
        A bigger bin number requires more data. In general,
        the larger the bin size, the closer the calibration error
        will be to the true calibration error.

    round_digits : int, default 4
        Round the calibration error metric.

    Returns
    -------
    calibration_error : float
        RMSE between the average positive label and predicted probability
        within each bin.
    """
    y_true = column_or_1d(y_true)
    y_prob = column_or_1d(y_prob)
    check_consistent_length(y_true, y_prob)

    binned_y_true, binned_y_prob = create_binned_data(y_true, y_prob, n_bins)

    # looping shouldn't be a source of bottleneck as n_bins should be a small number.
    bin_errors = 0.0
    for bin_y_true, bin_y_prob in zip(binned_y_true, binned_y_prob):
        avg_y_true = np.mean(bin_y_true)
        avg_y_score = np.mean(bin_y_prob)
        bin_error = (avg_y_score - avg_y_true) ** 2
        bin_errors += bin_error * len(bin_y_true)

    calibration_error = math.sqrt(bin_errors / len(y_true))
    return round(calibration_error, round_digits)


def create_binned_data(
    y_true: np.ndarray,
    y_prob: np.ndarray,
    n_bins: int) -> Tuple[List[np.ndarray], List[np.ndarray]]:
    """
    Bin ``y_true`` and ``y_prob`` by distribution of the data.
    i.e. each bin will contain approximately an equal number of
    data points. Bins are sorted based on ascending order of ``y_prob``.

    Parameters
    ----------
    y_true : 1d ndarray
        Binary true targets.

    y_prob : 1d ndarray
        Raw probability/score of the positive class.

    n_bins : int, default 15
        A bigger bin number requires more data.

    Returns
    -------
    binned_y_true/binned_y_prob : 1d ndarray
        Each element in the list stores the data for that bin.
    """
    sorted_indices = np.argsort(y_prob)
    sorted_y_true = y_true[sorted_indices]
    sorted_y_prob = y_prob[sorted_indices]
    binned_y_true = np.array_split(sorted_y_true, n_bins)
    binned_y_prob = np.array_split(sorted_y_prob, n_bins)
    return binned_y_true, binned_y_prob


def get_bin_boundaries(binned_y_prob: List[np.ndarray]) -> np.ndarray:
    """
    Given ``binned_y_prob`` from ``create_binned_data`` get the
    boundaries for each bin.

    Parameters
    ----------
    binned_y_prob : list
        Each element in the list stores the data for that bin.

    Returns
    -------
    bins : 1d ndarray
        Boundaries for each bin.
    """
    bins = []
    for i in range(len(binned_y_prob) - 1):
        last_prob = binned_y_prob[i][-1]
        next_first_prob = binned_y_prob[i + 1][0]
        bins.append((last_prob + next_first_prob) / 2.0)

    bins.append(1.0)
    return np.array(bins)


def compute_binary_score(
    y_true: np.ndarray,
    y_prob: np.ndarray,
    round_digits: int=4) -> Dict[str, float]:
    """
    Compute various evaluation metrics for binary classification.
    Including auc, precision, recall, f1, log loss, brier score. The
    threshold for precision and recall numbers are based on the one
    that gives the best f1 score.

    Parameters
    ----------
    y_true : 1d ndarray
        Binary true targets.

    y_prob : 1d ndarray
        Raw probability/score of the positive class.

    round_digits : int, default 4
        Round the evaluation metric.

    Returns
    -------
    metrics_dict : dict
        Metrics are stored in key value pair. ::

        {
            'auc': 0.82,
            'precision': 0.56,
            'recall': 0.61,
            'f1': 0.59,
            'log_loss': 0.42,
            'brier': 0.12
        }
    """
    auc = round(metrics.roc_auc_score(y_true, y_prob), round_digits)
    log_loss = round(metrics.log_loss(y_true, y_prob), round_digits)
    brier_score = round(metrics.brier_score_loss(y_true, y_prob), round_digits)

    precision, recall, threshold = metrics.precision_recall_curve(y_true, y_prob)
    f1 = 2 * (precision * recall) / (precision + recall)

    mask = ~np.isnan(f1)
    f1 = f1[mask]
    precision = precision[mask]
    recall = recall[mask]

    best_index = np.argmax(f1)
    precision = round(precision[best_index], round_digits)
    recall = round(recall[best_index], round_digits)
    f1 = round(f1[best_index], round_digits)
    return {
        'auc': auc,
        'precision': precision,
        'recall': recall,
        'f1': f1,
        'log_loss': log_loss,
        'brier': brier_score
    }


def compute_calibration_summary(
    eval_dict: Dict[str, pd.DataFrame],
    label_col: str='label',
    score_col: str='score',
    n_bins: int=15,
    strategy: str='quantile',
    round_digits: int=4,
    show: bool=True,
    save_plot_path: Optional[str]=None) -> pd.DataFrame:
    """
    Plots the calibration curve and computes the summary statistics for the model.

    Parameters
    ----------
    eval_dict : dict
        We can evaluate multiple calibration model's performance in one go. The key
        is the model name used to distinguish different calibration model, the value
        is the dataframe that stores the binary true targets and the predicted score
        for the positive class.

    label_col : str
        Column name for the dataframe in ``eval_dict`` that stores the binary true targets.

    score_col : str
        Column name for the dataframe in ``eval_dict`` that stores the predicted score.

    n_bins : int, default 15
        Number of bins to discretize the calibration curve plot and calibration error statistics.
        A bigger number requires more data, but will be closer to the true calibration error.

    strategy : {'uniform', 'quantile'}, default 'quantile'
        Strategy used to define the boundary of the bins.

        - uniform: The bins have identical widths.
        - quantile: The bins have the same number of samples and depend on the predicted score.

    round_digits : default 4
        Round the evaluation metric.

    show : bool, default True
        Whether to show the plots on the console or jupyter notebook.

    save_plot_path : str, default None
        Path where we'll store the calibration plot. None means it will not save the plot.

    Returns
    -------
    df_metrics : pd.DataFrame
        Corresponding metrics for all the input dataframe.
    """

    fig, (ax1, ax2) = plt.subplots(2)

    # estimator_metrics stores list of dict, e.g.
    # [{'auc': 0.776, 'name': 'xgb'}]
    estimator_metrics = []
    for name, df_eval in eval_dict.items():
        prob_true, prob_pred = calibration_curve(
            df_eval[label_col],
            df_eval[score_col],
            n_bins=n_bins,
            strategy=strategy)

        calibration_error = compute_calibration_error(
            df_eval[label_col], df_eval[score_col], n_bins, round_digits)
        metrics_dict = compute_binary_score(df_eval[label_col], df_eval[score_col], round_digits)
        metrics_dict['calibration_error'] = calibration_error
        metrics_dict['name'] = name
        estimator_metrics.append(metrics_dict)

        ax1.plot(prob_pred, prob_true, 's-', label=name)
        ax2.hist(df_eval[score_col], range=(0, 1), bins=n_bins, label=name, histtype='step', lw=2)

    ax1.plot([0, 1], [0, 1], 'k:', label='perfect')

    ax1.set_xlabel('Fraction of positives (Predicted)')
    ax1.set_ylabel('Fraction of positives (Actual)')
    ax1.set_ylim([-0.05, 1.05])
    ax1.legend(loc='upper left', ncol=2)
    ax1.set_title('Calibration Plots (Reliability Curve)')

    ax2.set_xlabel('Predicted scores')
    ax2.set_ylabel('Count')
    ax2.set_title('Histogram of Predicted Scores')
    ax2.legend(loc='upper right', ncol=2)

    plt.tight_layout()
    if show:
        plt.show()

    if save_plot_path is not None:
        save_dir = os.path.dirname(save_plot_path)
        if save_dir:
            os.makedirs(save_dir, exist_ok=True)

        fig.savefig(save_plot_path, dpi=300, bbox_inches='tight')

    plt.close(fig)

    df_metrics = pd.DataFrame(estimator_metrics)
    return df_metrics


import logging
import os

import matplotlib.pyplot as plt
import foldedleastsquares
import wotan
from lcbuilder.helper import LcbuilderHelper
from numpy.random import default_rng
from exoml.ete6.ete6_generator import Ete6ModelGenerator
from sklearn.utils import shuffle
import numpy as np
import pandas as pd


class WatsonPlanetModelGenerator(Ete6ModelGenerator):
    def __init__(self, injected_objects_df, lcs_dir, star_filename, lc_filename,
                 batch_size, input_sizes, transits_mask=None, zero_epsilon=1e-7,
                 measurements_per_point=1, plot_inputs=False, use_csvs=False, explain=False):
        super().__init__(zero_epsilon, shuffle_batch=False)
        self.injected_objects_df = injected_objects_df
        self.arrays_cache: dict = {}
        self.EXPLAIN_PERIOD_VALUE = [self.zero_epsilon, 0.5, 1 - self.zero_epsilon]
        self.EXPLAIN_DURATION_VALUE = [self.zero_epsilon, 0.5, 1 - self.zero_epsilon]
        self.EXPLAIN_RADIUS_VALUE = [self.zero_epsilon, 0.5, 1 - self.zero_epsilon]
        self.EXPLAIN_GTCN_VALUE = [self.zero_epsilon, 0.5, 1 - self.zero_epsilon]
        self.EXPLAIN_GTR_VALUE = [self.zero_epsilon, 0.5, 1 - self.zero_epsilon]
        self.EXPLAIN_OFFSET_VALUE = [self.zero_epsilon, 0.5, 1 - self.zero_epsilon]
        self.EXPLAIN_BFAP_VALUE = [self.zero_epsilon, 0.5, 1 - self.zero_epsilon]
        self.EXPLAIN_ODDEVEN_VALUE = [self.zero_epsilon, 0.5, 1 - self.zero_epsilon]
        self.EXPLAIN_TEMP_VALUE = [self.zero_epsilon, 0.5, 1 - self.zero_epsilon]
        self.EXPLAIN_ALBEDO_VALUE = [self.zero_epsilon, 0.5, 1 - self.zero_epsilon]
        self.EXPLAIN_STARTEFF_VALUE = [self.zero_epsilon, 0.5, 1 - self.zero_epsilon]
        self.EXPLAIN_STARRAD_VALUE = [self.zero_epsilon, 0.5, 1 - self.zero_epsilon]
        self.lcs_dir = lcs_dir
        self.star_filename = star_filename
        self.lc_filename = lc_filename
        self.batch_size = batch_size
        self.input_sizes = input_sizes
        self.transit_masks = transits_mask
        self.random_number_generator = default_rng()
        self.measurements_per_point = measurements_per_point
        self.plot_inputs = plot_inputs
        self.use_csvs = use_csvs
        self.explain = explain
        if self.explain:
            final_df = pd.DataFrame(
                columns=['object_id', 'period', 'duration(h)', 'depth_primary', 'radius(earth)', 'type',
                         'EXPLAIN_PERIOD_VALUE', 'EXPLAIN_DURATION_VALUE', 'EXPLAIN_RADIUS_VALUE', 'EXPLAIN_GTCN_VALUE',
                         'EXPLAIN_GTR_VALUE', 'EXPLAIN_OFFSET_VALUE', 'EXPLAIN_BFAP_VALUE',
                         'EXPLAIN_ODDEVEN_VALUE', 'EXPLAIN_TEMP_VALUE', 'EXPLAIN_ALBEDO_VALUE',
                         'EXPLAIN_STARTEFF_VALUE', 'EXPLAIN_STARRAD_VALUE',
                         'EXPLAIN_GLOBAL_VIEW', 'EXPLAIN_MAIN_VIEW', 'EXPLAIN_SECONDARY_VIEW',
                         'EXPLAIN_ODD_VIEW', 'EXPLAIN_EVEN_VIEW', 'EXPLAIN_CENTROIDS_RA_VIEW',
                         'EXPLAIN_CENTROIDS_DEC_VIEW', 'EXPLAIN_OG_VIEW'])
            for index, row in self.injected_objects_df.iterrows():
                row_copy_df = row.copy().to_frame().T
                final_df = pd.concat([final_df, row_copy_df], ignore_index=True)
                row_copy_df = row.copy().to_frame().T
                row_copy_df['EXPLAIN_GLOBAL_VIEW'] = True
                row_copy_df['object_id'] = row_copy_df['object_id'] + ' global view branch'
                final_df = pd.concat([final_df, row_copy_df], ignore_index=True)
                row_copy_df = row.copy().to_frame().T
                row_copy_df['EXPLAIN_MAIN_VIEW'] = True
                row_copy_df['object_id'] = row_copy_df['object_id'] + ' main view branch'
                final_df = pd.concat([final_df, row_copy_df], ignore_index=True)
                row_copy_df = row.copy().to_frame().T
                row_copy_df['EXPLAIN_SECONDARY_VIEW'] = True
                row_copy_df['object_id'] = row_copy_df['object_id'] + ' secondary view branch'
                final_df = pd.concat([final_df, row_copy_df], ignore_index=True)
                row_copy_df = row.copy().to_frame().T
                row_copy_df['EXPLAIN_ODD_VIEW'] = True
                row_copy_df['object_id'] = row_copy_df['object_id'] + ' odd view branch'
                final_df = pd.concat([final_df, row_copy_df], ignore_index=True)
                row_copy_df = row.copy().to_frame().T
                row_copy_df['EXPLAIN_EVEN_VIEW'] = True
                row_copy_df['object_id'] = row_copy_df['object_id'] + ' even view branch'
                final_df = pd.concat([final_df, row_copy_df], ignore_index=True)
                row_copy_df = row.copy().to_frame().T
                row_copy_df['EXPLAIN_CENTROIDS_RA_VIEW'] = True
                row_copy_df['object_id'] = row_copy_df['object_id'] + ' centroids ra branch'
                final_df = pd.concat([final_df, row_copy_df], ignore_index=True)
                row_copy_df = row.copy().to_frame().T
                row_copy_df['EXPLAIN_CENTROIDS_DEC_VIEW'] = True
                row_copy_df['object_id'] = row_copy_df['object_id'] + ' centroids dec branch'
                final_df = pd.concat([final_df, row_copy_df], ignore_index=True)
                row_copy_df = row.copy().to_frame().T
                row_copy_df['EXPLAIN_OG_VIEW'] = True
                row_copy_df['object_id'] = row_copy_df['object_id'] + ' og branch'
                final_df = pd.concat([final_df, row_copy_df], ignore_index=True)
                for value in self.EXPLAIN_PERIOD_VALUE:
                    row_copy_df = row.copy().to_frame().T
                    row_copy_df['EXPLAIN_PERIOD_VALUE'] = value
                    row_copy_df['object_id'] = row_copy_df['object_id'] + f' transit period value={value}'
                    final_df = pd.concat([final_df, row_copy_df], ignore_index=True)
                for value in self.EXPLAIN_DURATION_VALUE:
                    row_copy_df = row.copy().to_frame().T
                    row_copy_df['EXPLAIN_DURATION_VALUE'] = value
                    row_copy_df['object_id'] = row_copy_df['object_id'] + f' transit duration value={value}'
                    final_df = pd.concat([final_df, row_copy_df], ignore_index=True)
                for value in self.EXPLAIN_RADIUS_VALUE:
                    row_copy_df = row.copy().to_frame().T
                    row_copy_df['EXPLAIN_RADIUS_VALUE'] = value
                    row_copy_df['object_id'] = row_copy_df['object_id'] + f' planet radius value={value}'
                    final_df = pd.concat([final_df, row_copy_df], ignore_index=True)
                for value in self.EXPLAIN_GTCN_VALUE:
                    row_copy_df = row.copy().to_frame().T
                    row_copy_df['EXPLAIN_GTCN_VALUE'] = value
                    row_copy_df['object_id'] = row_copy_df['object_id'] + f' good transit count value={value}'
                    final_df = pd.concat([final_df, row_copy_df], ignore_index=True)
                for value in self.EXPLAIN_GTR_VALUE:
                    row_copy_df = row.copy().to_frame().T
                    row_copy_df['EXPLAIN_GTR_VALUE'] = value
                    row_copy_df['object_id'] = row_copy_df['object_id'] + f' good transit ratio value={value}'
                    final_df = pd.concat([final_df, row_copy_df], ignore_index=True)
                for value in self.EXPLAIN_OFFSET_VALUE:
                    row_copy_df = row.copy().to_frame().T
                    row_copy_df['EXPLAIN_OFFSET_VALUE'] = value
                    row_copy_df['object_id'] = row_copy_df['object_id'] + f' source offset value={value}'
                    final_df = pd.concat([final_df, row_copy_df], ignore_index=True)
                for value in self.EXPLAIN_BFAP_VALUE:
                    row_copy_df = row.copy().to_frame().T
                    row_copy_df['EXPLAIN_BFAP_VALUE'] = value
                    row_copy_df['object_id'] = row_copy_df['object_id'] + f' bootstrap fap value={value}'
                    final_df = pd.concat([final_df, row_copy_df], ignore_index=True)
                for value in self.EXPLAIN_ODDEVEN_VALUE:
                    row_copy_df = row.copy().to_frame().T
                    row_copy_df['EXPLAIN_ODDEVEN_VALUE'] = value
                    row_copy_df['object_id'] = row_copy_df['object_id'] + f' odd/even factor value={value}'
                    final_df = pd.concat([final_df, row_copy_df], ignore_index=True)
                for value in self.EXPLAIN_TEMP_VALUE:
                    row_copy_df = row.copy().to_frame().T
                    row_copy_df['EXPLAIN_TEMP_VALUE'] = value
                    row_copy_df['object_id'] = row_copy_df['object_id'] + f' temperature statistic value={value}'
                    final_df = pd.concat([final_df, row_copy_df], ignore_index=True)
                for value in self.EXPLAIN_ALBEDO_VALUE:
                    row_copy_df = row.copy().to_frame().T
                    row_copy_df['EXPLAIN_ALBEDO_VALUE'] = value
                    row_copy_df['object_id'] = row_copy_df['object_id'] + f' albedo statistic value={value}'
                    final_df = pd.concat([final_df, row_copy_df], ignore_index=True)
                for value in self.EXPLAIN_STARTEFF_VALUE:
                    row_copy_df = row.copy().to_frame().T
                    row_copy_df['EXPLAIN_STARTEFF_VALUE'] = value
                    row_copy_df['object_id'] = row_copy_df['object_id'] + f' star teff value={value}'
                    final_df = pd.concat([final_df, row_copy_df], ignore_index=True)
                for value in self.EXPLAIN_STARRAD_VALUE:
                    row_copy_df = row.copy().to_frame().T
                    row_copy_df['EXPLAIN_STARRAD_VALUE'] = value
                    row_copy_df['object_id'] = row_copy_df['object_id'] + f' star radius value={value}'
                    final_df = pd.concat([final_df, row_copy_df], ignore_index=True)
            self.injected_objects_df = final_df

    def __len__(self):
        return (np.ceil(len(self.injected_objects_df) / float(self.batch_size))).astype(int)

    def class_weights(self):
        return {0: 1, 1: 1}

    def _plot_df(self, df, type, scenario):
        fig, axs = plt.subplots(2, 3, figsize=(12, 6), constrained_layout=True)
        axs[0][0].scatter(df['time'], df['flux'])
        axs[0][1].scatter(df['time'], df['centroid_x'])
        axs[0][2].scatter(df['time'], df['centroid_y'])
        axs[1][0].scatter(df['time'], df['bck_flux'])
        axs[1][1].scatter(df['time'], df['motion_x'])
        axs[1][2].scatter(df['time'], df['motion_y'])
        plt.title(type + " " + scenario)
        plt.show()
        plt.clf()
        plt.close()

    def _plot_input(self, input_array, input_err_array, type, scenario, save_dir=None):
        if self.plot_inputs:
            transposed_err_array = []
            transposed_array = np.transpose(input_array)
            fig, axs = plt.subplots(2, 2, figsize=(24, 12), constrained_layout=True)
            current_array = transposed_array[0]
            current_array_mask = np.argwhere((~np.isnan(current_array)) & (current_array > self.zero_epsilon)).flatten()
            current_array = current_array[current_array_mask]
            time_array = current_array
            axs[0][0].scatter(np.arange(0, len(current_array)), current_array)
            # axs[0][0].set_ylim(np.mean(current_array) - 6 * np.std(current_array), np.mean(current_array) + 6 * np.std(current_array))
            axs[0][0].set_title("Time")
            current_array = transposed_array[1]
            current_array_mask = np.argwhere((~np.isnan(current_array)) & (current_array > self.zero_epsilon)).flatten()
            current_array = current_array[current_array_mask]
            if input_err_array is not None:
                transposed_err_array = np.transpose(input_err_array)
                axs[0][1].errorbar(time_array[current_array_mask], current_array, ls='none',
                                   yerr=transposed_err_array[1][current_array_mask], color="orange", alpha=0.5)
            axs[0][1].scatter(time_array[current_array_mask], current_array)
            if len(transposed_array) > 2:
                current_array = transposed_array[2]
                current_array_mask = np.argwhere(
                    (~np.isnan(current_array)) & (current_array > self.zero_epsilon)).flatten()
                current_array = current_array[current_array_mask]
                if input_err_array is not None:
                    axs[1][0].errorbar(time_array[current_array_mask], transposed_array[2], ls='none',
                                       yerr=transposed_err_array[2][current_array_mask], color="orange", alpha=0.5)
                axs[1][0].scatter(time_array[current_array_mask], current_array)
            #axs[0][1].set_ylim(np.mean(current_array) - 6 * np.std(current_array), np.mean(current_array) + 6 * np.std(current_array))
            axs[0][1].set_title("Flux")
            axs[1][0].set_title("Flux 1")
            # current_array = transposed_array[8]
            # current_array_mask = np.argwhere((~np.isnan(current_array)) & (current_array > self.zero_epsilon)).flatten()
            # current_array = current_array[current_array_mask]
            # axs[1][1].errorbar(time_array[current_array_mask], current_array,
            #                    yerr=transposed_err_array[8][current_array_mask], color="orange", alpha=0.5)
            # axs[1][1].scatter(time_array[current_array_mask], current_array)
            # # axs[1][1].set_ylim(np.mean(current_array) - 6 * np.std(current_array),
            # #                    np.mean(current_array) + 6 * np.std(current_array))
            # axs[1][1].set_title("Flux 2")
            # current_array = transposed_array[9]
            # current_array_mask = np.argwhere((~np.isnan(current_array)) & (current_array > self.zero_epsilon)).flatten()
            # current_array = current_array[current_array_mask]
            # axs[2][0].errorbar(time_array[current_array_mask], current_array,
            #                    yerr=transposed_err_array[9][current_array_mask], color="orange", alpha=0.5)
            # axs[2][0].scatter(time_array[current_array_mask], current_array)
            # # axs[2][0].set_ylim(np.mean(current_array) - 6 * np.std(current_array),
            # #                    np.mean(current_array) + 6 * np.std(current_array))
            # axs[2][0].set_title("Flux 3")
            # current_array = transposed_array[10]
            # current_array_mask = np.argwhere((~np.isnan(current_array)) & (current_array > self.zero_epsilon)).flatten()
            # current_array = current_array[current_array_mask]
            # axs[2][1].errorbar(time_array[current_array_mask], current_array,
            #                    yerr=transposed_err_array[10][current_array_mask], color="orange", alpha=0.5)
            # axs[2][1].scatter(time_array[current_array_mask], current_array)
            # # axs[2][1].set_ylim(np.mean(current_array) - 6 * np.std(current_array),
            # #                    np.mean(current_array) + 6 * np.std(current_array))
            # axs[2][1].set_title("Flux 4")
            # current_array = transposed_array[11]
            # current_array_mask = np.argwhere((~np.isnan(current_array)) & (current_array > self.zero_epsilon)).flatten()
            # current_array = current_array[current_array_mask]
            # axs[3][0].errorbar(time_array[current_array_mask], current_array,
            #                    yerr=transposed_err_array[11][current_array_mask], color="orange", alpha=0.5)
            # axs[3][0].scatter(time_array[current_array_mask], current_array)
            # # axs[3][0].set_ylim(np.mean(current_array) - 6 * np.std(current_array),
            # #                    np.mean(current_array) + 6 * np.std(current_array))
            # axs[3][0].set_title("Flux 5")
            # current_array = transposed_array[6]
            # current_array_mask = np.argwhere((~np.isnan(current_array)) & (current_array > self.zero_epsilon)).flatten()
            # current_array = current_array[current_array_mask]
            # axs[3][1].errorbar(time_array[current_array_mask], current_array,
            #                    yerr=transposed_err_array[6][current_array_mask], color="orange", alpha=0.5)
            # axs[3][1].scatter(time_array[current_array_mask], current_array)
            # # axs[3][1].set_ylim(np.mean(current_array) - 6 * np.std(current_array), np.mean(current_array) + 6 * np.std(current_array))
            # axs[3][1].set_title("Bck Flux")
            # current_array = transposed_array[2]
            # current_array_mask = np.argwhere((~np.isnan(current_array)) & (current_array > self.zero_epsilon)).flatten()
            # current_array = current_array[current_array_mask]
            # axs[4][0].errorbar(time_array[current_array_mask], current_array,
            #                    yerr=transposed_err_array[2][current_array_mask], color="orange", alpha=0.5)
            # axs[4][0].scatter(time_array[current_array_mask], current_array)
            # # axs[4][0].set_ylim(np.mean(current_array) - 6 * np.std(current_array), np.mean(current_array) + 6 * np.std(current_array))
            # axs[4][0].set_title("Centroid X")
            # current_array = transposed_array[3]
            # current_array_mask = np.argwhere((~np.isnan(current_array)) & (current_array > self.zero_epsilon)).flatten()
            # current_array = current_array[current_array_mask]
            # axs[4][1].errorbar(time_array[current_array_mask], current_array,
            #                    yerr=transposed_err_array[3][current_array_mask], color="orange", alpha=0.5)
            # axs[4][1].scatter(time_array[current_array_mask], current_array)
            # # axs[4][1].set_ylim(np.mean(current_array) - 6 * np.std(current_array), np.mean(current_array) + 6 * np.std(current_array))
            # axs[4][1].set_title("Centroid Y")
            # current_array = transposed_array[4]
            # current_array_mask = np.argwhere((~np.isnan(current_array)) & (current_array > self.zero_epsilon)).flatten()
            # current_array = current_array[current_array_mask]
            # axs[5][0].errorbar(time_array[current_array_mask], current_array,
            #                    yerr=transposed_err_array[4][current_array_mask], color="orange", alpha=0.5)
            # axs[5][0].scatter(time_array[current_array_mask], current_array)
            # # axs[5][0].set_ylim(np.mean(current_array) - 6 * np.std(current_array), np.mean(current_array) + 6 * np.std(current_array))
            # axs[5][0].set_title("Motion Y")
            # current_array = transposed_array[5]
            # current_array_mask = np.argwhere((~np.isnan(current_array)) & (current_array > self.zero_epsilon)).flatten()
            # current_array = current_array[current_array_mask]
            # axs[5][1].errorbar(time_array[current_array_mask], current_array,
            #                    yerr=transposed_err_array[5][current_array_mask], color="orange", alpha=0.5)
            # axs[5][1].scatter(time_array[current_array_mask], current_array)
            # # axs[5][1].set_ylim(np.mean(current_array) - 6 * np.std(current_array), np.mean(current_array) + 6 * np.std(current_array))
            # axs[5][1].set_title("Motion Y")
            fig.suptitle(type + " " + scenario)
            if save_dir:
                if not os.path.exists(save_dir):
                    os.mkdir(save_dir)
                plt.savefig(f'{save_dir}/{type}_{scenario}.png')
            else:
                plt.show()
            plt.clf()
            plt.close()

    def plot_single_data(self, lc_df, target_row):
        fig, axs = plt.subplots(1, 1, figsize=(8, 4), constrained_layout=True)
        axs.scatter(lc_df['time'], lc_df['flux_0'])
        axs.set_ylim(0.5 - target_row['tce_depth'] / 0.5e6, 0.505)
        plt.show()
        plt.clf()
        plt.close()

    def mask_other_signals(self, data_df, transits_mask, time_key='time'):
        if transits_mask is not None:
            for item_mask in transits_mask:
                mask = foldedleastsquares.transit_mask(data_df[time_key].to_numpy(), item_mask['P'],
                                                       2 * item_mask['D'] / 60 / 24, item_mask['T0'])
                data_df = data_df[~mask]
        return data_df

    def __getitem__(self, idx):
        max_index = (idx + 1) * self.batch_size
        max_index = len(self.injected_objects_df) if max_index > len(self.injected_objects_df) else max_index
        target_indexes = np.arange(idx * self.batch_size, max_index, 1)
        injected_objects_df = self.injected_objects_df.iloc[target_indexes]
        if self.shuffle_batch:
            injected_objects_df = shuffle(injected_objects_df)
        star_array = np.empty((len(target_indexes), 2, 1))
        star_neighbours_array = np.empty((len(target_indexes), self.input_sizes[1], 1))
        #[period, planet radius, number of transits, ratio of good transits, transit depth, transit_offset_pos - transit_offset_err]
        scalar_values = np.empty((len(target_indexes), 12, 1))
        global_flux_array = np.empty((len(target_indexes), self.input_sizes[2], self.measurements_per_point))
        folded_flux_even_array = np.empty((len(target_indexes), self.input_sizes[3], self.measurements_per_point))
        folded_flux_odd_array = np.empty((len(target_indexes), self.input_sizes[3], self.measurements_per_point))
        folded_flux_even_subhar_array = np.empty((len(target_indexes), self.input_sizes[3], self.measurements_per_point))
        folded_flux_odd_subhar_array = np.empty((len(target_indexes), self.input_sizes[3], self.measurements_per_point))
        folded_flux_even_har_array = np.empty((len(target_indexes), self.input_sizes[3], self.measurements_per_point))
        folded_flux_odd_har_array = np.empty((len(target_indexes), self.input_sizes[3], self.measurements_per_point))
        global_flux_array_err = np.empty((len(target_indexes), self.input_sizes[2], self.measurements_per_point))
        folded_flux_even_array_err = np.empty((len(target_indexes), self.input_sizes[3], self.measurements_per_point))
        folded_flux_odd_array_err = np.empty((len(target_indexes), self.input_sizes[3], self.measurements_per_point))
        folded_flux_even_subhar_array_err = np.empty((len(target_indexes), self.input_sizes[3], self.measurements_per_point))
        folded_flux_odd_subhar_array_err = np.empty((len(target_indexes), self.input_sizes[3], self.measurements_per_point))
        folded_flux_even_har_array_err = np.empty((len(target_indexes), self.input_sizes[3], self.measurements_per_point))
        folded_flux_odd_har_array_err = np.empty((len(target_indexes), self.input_sizes[3], self.measurements_per_point))
        folded_centroids_array = np.empty((len(target_indexes), self.input_sizes[3], 2))
        folded_centroids_array_err = np.empty((len(target_indexes), self.input_sizes[3], 2))
        folded_og_array = np.empty((len(target_indexes), self.input_sizes[3], 1))
        folded_og_array_err = np.empty((len(target_indexes), self.input_sizes[3], 1))
        batch_data_values = np.empty((len(target_indexes), 1))
        i = 0
        for df_index, target_row in injected_objects_df.iterrows():
            #TODO refactor to use mission for object_id
            object_id = target_row['object_id'].split(' ')
            mission_id = object_id[0]
            target_id = int(object_id[1].split('_')[0])
            period = target_row['period']
            epoch = target_row['epoch']
            duration = target_row['duration(h)'] / 24
            duration_to_period = duration / period
            batch_data_values[i] = [0]
            file_prefix = self.lcs_dir + '/' #+ mission_id + '_' + str(target_id)
            inputs_save_dir = f'{self.lcs_dir}/iatson/'
            if (not self.use_csvs and not self.explain) or (self.explain and df_index == 0):
                if target_row['object_id'] in self.arrays_cache and 'folded_og_array' in self.arrays_cache[target_row['object_id']]:
                    folded_og_array[i] = self.arrays_cache[target_row['object_id']]['folded_og_array']
                else:
                    og_filename = file_prefix + 'og_dg.csv'
                    og_df = pd.read_csv(og_filename, usecols=['time', 'og_flux', 'halo_flux', 'core_flux'], low_memory=True)
                    og_df = self.mask_other_signals(og_df, transits_mask=self.transit_masks, time_key='time')
                    og_df = self.compute_og_df(og_df, object_id, duration)
                    og_df = og_df.sort_values(by=['time'])
                    og_df['time'] = self.fold(og_df['time'].to_numpy(), period, epoch + period / 2)
                    og_df = og_df.sort_values(by=['time'])
                    og_df = self._prepare_input_og(og_df)
                    og_df = og_df[(og_df['time'] > 0.5 - duration_to_period * 3) & (
                            og_df['time'] < 0.5 + duration_to_period * 3)]
                    folded_og_array[i], folded_og_array_err[i] = self.bin_by_time(og_df.to_numpy(), self.input_sizes[3],
                                                                                  target_row['object_id'])
                    self.arrays_cache[target_row['object_id']] = {}
                    self.arrays_cache[target_row['object_id']]['folded_og_array'] = folded_og_array[i]
                if target_id in self.arrays_cache and 'folded_centroids_array' in self.arrays_cache[target_row['object_id']]:
                    folded_centroids_array[i] = self.arrays_cache[target_row['object_id']]['folded_centroids_array']
                else:
                    centroids_filename = file_prefix + 'centroids.csv'
                    centroids_df = pd.read_csv(centroids_filename, usecols=['time', 'centroids_ra', 'centroids_dec'],
                                                        low_memory=True)
                    centroids_df = self.mask_other_signals(centroids_df, transits_mask=self.transit_masks, time_key='time')
                    centroids_df = centroids_df.sort_values(by=['time'])
                    centroids_df['centroids_ra'], _ = LcbuilderHelper.detrend(centroids_df['time'].to_numpy(),
                                                                                   centroids_df['centroids_ra'].to_numpy(),
                                                                                   duration * 4,
                                                                                   check_cadence=True)
                    centroids_df['centroids_dec'], _ = LcbuilderHelper.detrend(
                        centroids_df['time'].to_numpy(),
                        centroids_df['centroids_dec'].to_numpy(), duration * 4,
                        check_cadence=True)
                    centroids_df['time'] = self.fold(centroids_df['time'].to_numpy(), period, epoch + period / 2)
                    centroids_df = centroids_df.sort_values(by=['time'])
                    centroids_df = self._prepare_input_centroids(centroids_df)
                    centroids_df = centroids_df[(centroids_df['time'] > 0.5 - duration_to_period * 3) & (
                            centroids_df['time'] < 0.5 + duration_to_period * 3)]
                    folded_centroids_array[i], folded_centroids_array_err[i] = self.bin_by_time(centroids_df.to_numpy(), self.input_sizes[3], target_row['object_id'])
                    self.arrays_cache[target_row['object_id']]['folded_centroids_array'] = folded_centroids_array[i]
                if (target_row['object_id'] in self.arrays_cache and 'star_teff' in self.arrays_cache[target_row['object_id']]) and \
                        (target_row['object_id'] in self.arrays_cache and 'star_radius' in self.arrays_cache[target_row['object_id']]):
                    star_array[i] = [[self.arrays_cache[target_row['object_id']]['star_teff']],
                                     [self.arrays_cache[target_row['object_id']]['star_radius']]]
                    ra = self.arrays_cache[target_row['object_id']]['ra']
                    dec = self.arrays_cache[target_row['object_id']]['dec']
                else:
                    star_df = pd.read_csv(self.star_filename, usecols=['Teff', 'radius', 'ra', 'dec'],
                                          index_col=False)
                    star_df, ra, dec = self._prepare_input_star(target_row, star_df)
                    star_df['Teff'] = star_df['Teff'] if star_df['Teff'] <= 1 else 1
                    star_df['radius'] = star_df['radius'] if star_df['radius'] <= 1 else 1
                    star_array[i] = np.transpose([star_df.to_numpy()])
                    self.arrays_cache[target_row['object_id']]['ra'] = ra
                    self.arrays_cache[target_row['object_id']]['dec'] = dec
                    self.arrays_cache[target_row['object_id']]['star_teff'] = star_df['Teff']
                    self.arrays_cache[target_row['object_id']]['star_radius'] = star_df['radius']
                if target_row['object_id'] in self.arrays_cache and 'scalar_values' in self.arrays_cache[target_row['object_id']]:
                    scalar_values[i] = self.arrays_cache[target_row['object_id']]['scalar_values']
                    global_flux_array[i] = self.arrays_cache[target_row['object_id']]['global_flux_array']
                    folded_flux_odd_array[i] = self.arrays_cache[target_row['object_id']]['folded_flux_odd_array']
                    folded_flux_even_array[i] = self.arrays_cache[target_row['object_id']]['folded_flux_even_array']
                    folded_flux_even_har_array[i] = self.arrays_cache[target_row['object_id']]['folded_flux_even_har_array']
                    folded_flux_odd_har_array[i] = self.arrays_cache[target_row['object_id']]['folded_flux_odd_har_array']
                    folded_flux_even_subhar_array[i] = self.arrays_cache[target_row['object_id']]['folded_flux_even_subhar_array']
                    folded_flux_odd_subhar_array[i] = self.arrays_cache[target_row['object_id']]['folded_flux_odd_subhar_array']
                else:
                    lc_filename = self.lc_filename
                    lc_df = pd.read_csv(lc_filename, usecols=['time', 'flux'], low_memory=True)
                    if lc_df is None:
                        logging.warning("No curve for target " + file_prefix)
                        raise ValueError("No curve for target " + file_prefix)
                    lc_df = lc_df.sort_values(by=['time'])
                    lc_df = self.mask_other_signals(lc_df, transits_mask=self.transit_masks, time_key='time')
                    lc_df['flux'] = wotan.flatten(lc_df['time'].to_numpy(), lc_df['flux'].to_numpy(), duration * 4,
                                                  method='biweight')
                    lc_df, good_transits_count, transits_count = self._prepare_input_lc(lc_df, period, epoch, duration,
                                                                                        time_key='time', flux_key='flux')
                    not_null_times_args = np.argwhere(lc_df['time'].to_numpy() > 0).flatten()
                    lc_df = lc_df.iloc[not_null_times_args]
                    time = lc_df['time'].to_numpy()
                    # Global flux
                    # Shifting data 1/4 so that main transit and possible occultation don't get cut by the borders
                    lc_df_sorted_fold = lc_df.copy()
                    lc_df_sorted_fold['time'] = self.fold(time, period, epoch + period / 4)
                    lc_df_sorted_fold = lc_df_sorted_fold.sort_values(by=['time'])
                    global_flux_array[i], global_flux_array_err[i] = \
                        self.bin_by_time(lc_df_sorted_fold.to_numpy(), self.input_sizes[2], target_row['object_id'])
                    # Focus flux even
                    lc_df_focus = lc_df.copy()
                    lc_df_focus['time'] = self.fold(time, period, epoch)
                    lc_df_focus = lc_df_focus.sort_values(by=['time'])
                    lc_df_focus = lc_df_focus[(lc_df_focus['time'] > 0.5 - duration_to_period * 3) &
                                              (lc_df_focus['time'] < 0.5 + duration_to_period * 3)]
                    folded_flux_even_array[i], folded_flux_even_array_err[i] = \
                        self.bin_by_time(lc_df_focus.to_numpy(), self.input_sizes[3], target_row['object_id'])
                    # Focus flux odd
                    lc_df_focus = lc_df.copy()
                    lc_df_focus['time'] = self.fold(time, period, epoch + period / 2)
                    lc_df_focus = lc_df_focus.sort_values(by=['#time'])
                    lc_df_focus = lc_df_focus[(lc_df_focus['#time'] > 0.5 - duration_to_period * 3) &
                                              (lc_df_focus['#time'] < 0.5 + duration_to_period * 3)]
                    folded_flux_odd_array[i], folded_flux_odd_array_err[i] = \
                        self.bin_by_time(lc_df_focus.to_numpy(), self.input_sizes[4], target_row['object_id'])
                    # Focus flux harmonic even
                    lc_df_focus = lc_df.copy()
                    lc_df_focus['#time'] = self.fold(time, period * 2, epoch)
                    lc_df_focus = lc_df_focus.sort_values(by=['#time'])
                    lc_df_focus = lc_df_focus[(lc_df_focus['#time'] > 0.5 - duration_to_period * 3) &
                                              (lc_df_focus['#time'] < 0.5 + duration_to_period * 3)]
                    oot_lc_df_focus = lc_df_focus[((lc_df_focus['#time'] <= 0.5 - duration_to_period / 4) |
                                                   (lc_df_focus['#time'] >= 0.5 + duration_to_period / 4))]
                    it_lc_df_focus = lc_df_focus[((lc_df_focus['#time'] > 0.5 - duration_to_period / 4) &
                                                  (lc_df_focus['#time'] < 0.5 + duration_to_period / 4))]
                    depth_even = (oot_lc_df_focus['flux'].mean() - it_lc_df_focus['flux'].mean()) * np.sqrt(
                        len(it_lc_df_focus)) / oot_lc_df_focus['flux'].std()
                    folded_flux_even_har_array[i], folded_flux_even_har_array_err[i] = \
                        self.bin_by_time(lc_df_focus.to_numpy(), self.input_sizes[7], target_row['object_id'])
                    # Focus flux harmonic odd
                    lc_df_focus = lc_df.copy()
                    lc_df_focus['#time'] = self.fold(time, period * 2, epoch + period)
                    lc_df_focus = lc_df_focus.sort_values(by=['#time'])
                    lc_df_focus = lc_df_focus[(lc_df_focus['#time'] > 0.5 - duration_to_period * 3) &
                                              (lc_df_focus['#time'] < 0.5 + duration_to_period * 3)]
                    oot_lc_df_focus = lc_df_focus[((lc_df_focus['#time'] <= 0.5 - duration_to_period / 4) |
                                                   (lc_df_focus['#time'] >= 0.5 + duration_to_period / 4))]
                    it_lc_df_focus = lc_df_focus[((lc_df_focus['#time'] > 0.5 - duration_to_period / 4) &
                                                  (lc_df_focus['#time'] < 0.5 + duration_to_period / 4))]
                    depth_odd = (oot_lc_df_focus['flux'].mean() - it_lc_df_focus['flux'].mean()) * np.sqrt(
                        len(it_lc_df_focus)) / oot_lc_df_focus['flux'].std()
                    folded_flux_odd_har_array[i], folded_flux_odd_har_array_err[i] = \
                        self.bin_by_time(lc_df_focus.to_numpy(), self.input_sizes[8], target_row['object_id'])
                    # Focus flux sub-harmonic even
                    lc_df_focus = pd.DataFrame(columns=['#time', 'flux'])
                    time, flux0, _ = LcbuilderHelper.mask_transits(time,
                                                                   lc_df.copy()['flux'].to_numpy(), period,
                                                                   duration * 2,
                                                                   epoch)
                    lc_df_focus['#time'] = time
                    lc_df_focus['flux'] = flux0
                    lc_df_focus['#time'] = self.fold(time, period / 2, epoch)
                    lc_df_focus = lc_df_focus.sort_values(by=['#time'])
                    lc_df_focus = lc_df_focus[(lc_df_focus['#time'] > 0.5 - duration_to_period * 3) &
                                              (lc_df_focus['#time'] < 0.5 + duration_to_period * 3)]
                    folded_flux_even_subhar_array[i], folded_flux_even_subhar_array_err[i] = \
                        self.bin_by_time(lc_df_focus.to_numpy(), self.input_sizes[5], target_row['object_id'])
                    # Focus flux sub-harmonic odd
                    lc_df_focus = pd.DataFrame(columns=['#time', 'flux'])
                    lc_df_focus['#time'] = time
                    lc_df_focus['flux'] = flux0
                    lc_df_focus['#time'] = self.fold(time, period / 2, epoch + period / 4)
                    lc_df_focus = lc_df_focus.sort_values(by=['#time'])
                    lc_df_focus = lc_df_focus[(lc_df_focus['#time'] > 0.5 - duration_to_period * 3) &
                                              (lc_df_focus['#time'] < 0.5 + duration_to_period * 3)]
                    folded_flux_odd_subhar_array[i], folded_flux_odd_subhar_array_err[i] = \
                        self.bin_by_time(lc_df_focus.to_numpy(), self.input_sizes[6], target_row['object_id'])
                    oe_factor = np.abs(depth_even - depth_odd)
                    oe_factor = oe_factor / 5 if oe_factor < 5 else 1 - self.zero_epsilon
                    oe_factor = oe_factor if not np.isnan(oe_factor) and oe_factor > 0 else self.zero_epsilon
                    offset_filename = file_prefix + 'source_offsets.csv'
                    offset_df = pd.read_csv(offset_filename, low_memory=True)
                    row = offset_df[offset_df['name'] == 'mean'].iloc[0]
                    offset_ra = row['ra']
                    offset_dec = row['dec']
                    offset_ra_err = row['ra_err']
                    offset_dec_err = row['dec_err']
                    target_dist = np.sqrt((offset_ra - ra) ** 2 + (offset_dec - dec) ** 2)
                    target_dist = target_dist if not np.isnan(target_dist) else self.zero_epsilon
                    offset_err = offset_ra_err if offset_ra_err > offset_dec_err else offset_dec_err
                    offset_err = offset_err if offset_err > 0 else target_dist * 2
                    offset_err = offset_err / (target_dist * 2)
                    target_dist = target_dist if target_dist < 1 else 1 - self.zero_epsilon
                    offset_err = 1 - self.zero_epsilon if offset_err >= 1 else offset_err
                    offset_err = offset_err if offset_err > 0 else self.zero_epsilon
                    metrics_df = pd.read_csv(file_prefix + 'metrics.csv')
                    temperature_stat = metrics_df.loc[metrics_df['metric'] == 'temp_stat'].iloc[0]['score']
                    albedo_stat = metrics_df.loc[metrics_df['metric'] == 'albedo_stat'].iloc[0]['score']
                    bootstrap_fap = metrics_df.loc[metrics_df['metric'] == 'bootstrap_fap'].iloc[0]['score']
                    oe_factor = np.abs(depth_even - depth_odd)
                    oe_factor = oe_factor / 5 if oe_factor < 5 else 1 - self.zero_epsilon
                    oe_factor = oe_factor if not np.isnan(oe_factor) and oe_factor > 0 else self.zero_epsilon
                    if np.isnan(temperature_stat):
                        temperature_stat = 0
                    elif temperature_stat < -10:
                        temperature_stat = -10
                    elif temperature_stat > 10:
                        temperature_stat = 10
                    temperature_stat = (temperature_stat + 10) / 20
                    temperature_stat = temperature_stat if temperature_stat > 0 else self.zero_epsilon
                    temperature_stat = temperature_stat if temperature_stat < 1 else 1 - self.zero_epsilon
                    if np.isnan(albedo_stat):
                        albedo_stat = 0
                    elif albedo_stat < -10:
                        albedo_stat = -10
                    elif albedo_stat > 10:
                        albedo_stat = 10
                    albedo_stat = (albedo_stat + 10) / 20
                    albedo_stat = albedo_stat if albedo_stat > 0 else self.zero_epsilon
                    albedo_stat = albedo_stat if albedo_stat < 1 else 1 - self.zero_epsilon
                    good_transits_count_norm = good_transits_count / 20
                    good_transits_count_norm = good_transits_count_norm if good_transits_count_norm < 1 else 1 - self.zero_epsilon
                    good_transits_ratio = good_transits_count / transits_count if transits_count > 0 else self.zero_epsilon
                    good_transits_ratio = good_transits_ratio if good_transits_ratio < 1 else 1 - self.zero_epsilon
                    planet_radius = target_row['radius(earth)'] / 300
                    planet_radius = planet_radius if planet_radius < 1 else 1 - self.zero_epsilon
                    depth = target_row['depth_primary'] / 1e3
                    depth = depth if depth < 1 else 1 - self.zero_epsilon
                    duration = duration / 15
                    duration = duration if duration < 1 else 1 - self.zero_epsilon
                    scalar_values[i] = np.transpose([[period / 1200 if period < 1200 else 1,
                                                      duration,
                                                      depth,
                                                      planet_radius,
                                                      good_transits_count_norm,
                                                      good_transits_ratio,
                                                      target_dist if not np.isnan(target_dist) else self.zero_epsilon,
                                                      offset_err,
                                                      bootstrap_fap,
                                                      oe_factor,
                                                      temperature_stat,
                                                      albedo_stat]])
                    self.arrays_cache[target_row['object_id']]['global_flux_array'] = global_flux_array[i]
                    self.arrays_cache[target_row['object_id']]['folded_flux_even_array'] = folded_flux_even_array[i]
                    self.arrays_cache[target_row['object_id']]['folded_flux_odd_array'] = folded_flux_odd_array[i]
                    self.arrays_cache[target_row['object_id']]['folded_flux_even_subhar_array'] = folded_flux_even_subhar_array[i]
                    self.arrays_cache[target_row['object_id']]['folded_flux_odd_subhar_array'] = folded_flux_odd_subhar_array[i]
                    self.arrays_cache[target_row['object_id']]['folded_flux_even_har_array'] = folded_flux_even_har_array[i]
                    self.arrays_cache[target_row['object_id']]['folded_flux_odd_har_array'] = folded_flux_odd_har_array[i]
                    self.arrays_cache[target_row['object_id']]['scalar_values'] = scalar_values[i]
            else:
                scalar_values[i] = [[e] for e in np.loadtxt(inputs_save_dir + 'input_scalar_values.csv', delimiter=',')]
                star_array[i] = [[e] for e in np.loadtxt(inputs_save_dir + 'input_star.csv', delimiter=',')]
                star_neighbours_array[i] = [[e] for e in np.loadtxt(inputs_save_dir + 'input_nb.csv', delimiter=',')]
                global_flux_array[i] = np.loadtxt(inputs_save_dir + 'input_global.csv', delimiter=',').reshape(-1, 1)
                global_flux_array_err[i] = np.loadtxt(inputs_save_dir + 'input_global_err.csv', delimiter=',').reshape(-1, 1)
                folded_flux_even_array[i] = np.loadtxt(inputs_save_dir + 'input_even.csv', delimiter=',').reshape(-1, 1)
                folded_flux_even_array_err[i] = np.loadtxt(inputs_save_dir + 'input_even_err.csv', delimiter=',').reshape(-1, 1)
                folded_flux_odd_array[i] = np.loadtxt(inputs_save_dir + 'input_odd.csv', delimiter=',').reshape(-1, 1)
                folded_flux_odd_array_err[i] = np.loadtxt(inputs_save_dir + 'input_odd_err.csv', delimiter=',').reshape(-1, 1)
                folded_flux_even_subhar_array[i] = np.loadtxt(inputs_save_dir + 'input_even_sh.csv', delimiter=',').reshape(-1, 1)
                folded_flux_even_subhar_array_err[i] = np.loadtxt(inputs_save_dir + 'input_even_sh_err.csv', delimiter=',').reshape(-1, 1)
                folded_flux_odd_subhar_array[i] = np.loadtxt(inputs_save_dir + 'input_odd_sh.csv', delimiter=',').reshape(-1, 1)
                folded_flux_odd_subhar_array_err[i] = np.loadtxt(inputs_save_dir + 'input_odd_sh_err.csv', delimiter=',').reshape(-1, 1)
                folded_flux_even_har_array[i] = np.loadtxt(inputs_save_dir + 'input_even_h.csv', delimiter=',').reshape(-1, 1)
                folded_flux_even_har_array_err[i] = np.loadtxt(inputs_save_dir + 'input_even_h_err.csv', delimiter=',').reshape(-1, 1)
                folded_flux_odd_har_array[i] = np.loadtxt(inputs_save_dir + 'input_odd_h.csv', delimiter=',').reshape(-1, 1)
                folded_flux_odd_har_array_err[i] = np.loadtxt(inputs_save_dir + 'input_odd_h_err.csv', delimiter=',').reshape(-1, 1)
                folded_centroids_array[i] = np.loadtxt(inputs_save_dir + 'input_centroids.csv', delimiter=',')
                folded_centroids_array_err[i] = np.loadtxt(inputs_save_dir + 'input_centroids_err.csv', delimiter=',')
                folded_og_array[i] = np.loadtxt(inputs_save_dir + 'input_og.csv', delimiter=',').reshape(-1, 1)
                folded_og_array_err[i] = np.loadtxt(inputs_save_dir + 'input_og_err.csv', delimiter=',').reshape(-1, 1)
            self.assert_in_range(object_id, scalar_values[i], None)
            self.assert_in_range(object_id, global_flux_array[i], None)
            self.assert_in_range(object_id, star_array[i], None)
            self.assert_in_range(object_id, folded_flux_even_array[i], None)
            self.assert_in_range(object_id, folded_flux_odd_array[i], None)
            self.assert_in_range(object_id, folded_flux_even_subhar_array[i], None)
            self.assert_in_range(object_id, folded_flux_odd_subhar_array[i], None)
            self.assert_in_range(object_id, folded_flux_even_har_array[i], None)
            self.assert_in_range(object_id, folded_flux_odd_har_array[i], None)
            self.assert_in_range(object_id, folded_centroids_array[i], None)
            self.assert_in_range(object_id, folded_og_array[i], None)
            if df_index == 0:
                self._plot_input(global_flux_array[i], None, target_row['object_id'], "global",
                                 save_dir=inputs_save_dir)
                self._plot_input(folded_flux_even_array[i], None, target_row['object_id'], "even",
                                 save_dir=inputs_save_dir)
                self._plot_input(folded_flux_odd_array[i], None, target_row['object_id'], "odd",
                                 save_dir=inputs_save_dir)
                self._plot_input(folded_flux_even_har_array[i], None, target_row['object_id'], "even_har",
                                 save_dir=inputs_save_dir)
                self._plot_input(folded_flux_odd_har_array[i], None, target_row['object_id'], "odd_har",
                                 save_dir=inputs_save_dir)
                # self._plot_input(folded_flux_even_subhar_array[i], folded_flux_even_subhar_array_err[i], target_row['object_id'], "even_subhar")
                # self._plot_input(folded_flux_odd_subhar_array[i], folded_flux_odd_subhar_array_err[i], target_row['object_id'], "odd_subhar")
                self._plot_input(folded_og_array[i], None, target_row['object_id'], "OG", save_dir=inputs_save_dir)
                self._plot_input(folded_centroids_array[i], None, target_row['object_id'], "CENTROIDS",
                                 save_dir=inputs_save_dir)
                np.savetxt(inputs_save_dir + 'input_scalar_values.csv', scalar_values[i], delimiter=',')
                np.savetxt(inputs_save_dir + 'input_star.csv', star_array[i], delimiter=',')
                np.savetxt(inputs_save_dir + 'input_nb.csv', star_neighbours_array[i], delimiter=',')
                np.savetxt(inputs_save_dir + 'input_global.csv', global_flux_array[i], delimiter=',')
                np.savetxt(inputs_save_dir + 'input_global_err.csv', global_flux_array_err[i], delimiter=',')
                np.savetxt(inputs_save_dir + 'input_even.csv', folded_flux_even_array[i], delimiter=',')
                np.savetxt(inputs_save_dir + 'input_even_err.csv', folded_flux_even_array_err[i], delimiter=',')
                np.savetxt(inputs_save_dir + 'input_odd.csv', folded_flux_odd_array[i], delimiter=',')
                np.savetxt(inputs_save_dir + 'input_odd_err.csv', folded_flux_odd_array_err[i], delimiter=',')
                np.savetxt(inputs_save_dir + 'input_even_sh.csv', folded_flux_even_subhar_array[i], delimiter=',')
                np.savetxt(inputs_save_dir + 'input_even_sh_err.csv', folded_flux_even_subhar_array_err[i], delimiter=',')
                np.savetxt(inputs_save_dir + 'input_odd_sh.csv', folded_flux_odd_subhar_array[i], delimiter=',')
                np.savetxt(inputs_save_dir + 'input_odd_sh_err.csv', folded_flux_odd_subhar_array_err[i], delimiter=',')
                np.savetxt(inputs_save_dir + 'input_even_h.csv', folded_flux_even_har_array[i], delimiter=',')
                np.savetxt(inputs_save_dir + 'input_even_h_err.csv', folded_flux_even_har_array_err[i], delimiter=',')
                np.savetxt(inputs_save_dir + 'input_odd_h.csv', folded_flux_odd_har_array[i], delimiter=',')
                np.savetxt(inputs_save_dir + 'input_odd_h_err.csv', folded_flux_odd_har_array_err[i], delimiter=',')
                np.savetxt(inputs_save_dir + 'input_centroids.csv', folded_centroids_array[i], delimiter=',')
                np.savetxt(inputs_save_dir + 'input_centroids_err.csv', folded_centroids_array_err[i], delimiter=',')
                np.savetxt(inputs_save_dir + 'input_og.csv', folded_og_array[i], delimiter=',')
                np.savetxt(inputs_save_dir + 'input_og_err.csv', folded_og_array_err[i], delimiter=',')
            if self.explain:
                if 'EXPLAIN_PERIOD_VALUE' in target_row and not np.isnan(target_row['EXPLAIN_PERIOD_VALUE']):
                    scalar_values[i, 0] = target_row['EXPLAIN_PERIOD_VALUE']
                if 'EXPLAIN_DURATION_VALUE' in target_row and not np.isnan(target_row['EXPLAIN_DURATION_VALUE']):
                    scalar_values[i, 1] = target_row['EXPLAIN_DURATION_VALUE']
                if 'EXPLAIN_RADIUS_VALUE' in target_row and not np.isnan(target_row['EXPLAIN_RADIUS_VALUE']):
                    scalar_values[i, 3] = target_row['EXPLAIN_RADIUS_VALUE']
                if 'EXPLAIN_GTCN_VALUE' in target_row and not np.isnan(target_row['EXPLAIN_GTCN_VALUE']):
                    scalar_values[i, 4] = target_row['EXPLAIN_GTCN_VALUE']
                if 'EXPLAIN_GTR_VALUE' in target_row and not np.isnan(target_row['EXPLAIN_GTR_VALUE']):
                    scalar_values[i, 5] = target_row['EXPLAIN_GTR_VALUE']
                if 'EXPLAIN_OFFSET_VALUE' in target_row and not np.isnan(target_row['EXPLAIN_OFFSET_VALUE']):
                    scalar_values[i, 7] = target_row['EXPLAIN_OFFSET_VALUE']
                if 'EXPLAIN_ODDEVEN_VALUE' in target_row and not np.isnan(target_row['EXPLAIN_ODDEVEN_VALUE']):
                    scalar_values[i, 8] = target_row['EXPLAIN_ODDEVEN_VALUE']
                if 'EXPLAIN_BFAP_VALUE' in target_row and not np.isnan(target_row['EXPLAIN_BFAP_VALUE']):
                    scalar_values[i, 9] = target_row['EXPLAIN_BFAP_VALUE']
                if 'EXPLAIN_TEMP_VALUE' in target_row and not np.isnan(target_row['EXPLAIN_TEMP_VALUE']):
                    scalar_values[i, 10] = target_row['EXPLAIN_TEMP_VALUE']
                if 'EXPLAIN_ALBEDO_VALUE' in target_row and not np.isnan(target_row['EXPLAIN_ALBEDO_VALUE']):
                    scalar_values[i, 11] = target_row['EXPLAIN_ALBEDO_VALUE']
                if 'EXPLAIN_STARTEFF_VALUE' in target_row and not np.isnan(target_row['EXPLAIN_STARTEFF_VALUE']):
                    star_array[i, 0] = target_row['EXPLAIN_STARTEFF_VALUE']
                if 'EXPLAIN_STARRAD_VALUE' in target_row and not np.isnan(target_row['EXPLAIN_STARRAD_VALUE']):
                    star_array[i, 1] = target_row['EXPLAIN_STARRAD_VALUE']
                if 'EXPLAIN_GLOBAL_VIEW' in target_row and not np.isnan(target_row['EXPLAIN_GLOBAL_VIEW']):
                    global_flux_array[i] = np.zeros(global_flux_array[i].shape)
                if 'EXPLAIN_MAIN_VIEW' in target_row and not np.isnan(target_row['EXPLAIN_MAIN_VIEW']):
                    folded_flux_odd_array[i] = np.zeros(folded_flux_odd_array[i].shape)
                if 'EXPLAIN_SECONDARY_VIEW' in target_row and not np.isnan(target_row['EXPLAIN_SECONDARY_VIEW']):
                    folded_flux_even_array[i] = np.zeros(folded_flux_even_array[i].shape)
                if 'EXPLAIN_ODD_VIEW' in target_row and not np.isnan(target_row['EXPLAIN_ODD_VIEW']):
                    folded_flux_odd_har_array[i] = np.zeros(folded_flux_odd_har_array[i].shape)
                if 'EXPLAIN_EVEN_VIEW' in target_row and not np.isnan(target_row['EXPLAIN_EVEN_VIEW']):
                    folded_flux_even_har_array[i] = np.zeros(folded_flux_even_har_array[i].shape)
                if 'EXPLAIN_CENTROIDS_RA_VIEW' in target_row and not np.isnan(target_row['EXPLAIN_CENTROIDS_RA_VIEW']):
                    folded_centroids_array[i, :, 0] = np.zeros(folded_centroids_array[i].shape[0])
                if 'EXPLAIN_CENTROIDS_DEC_VIEW' in target_row and not np.isnan(target_row['EXPLAIN_CENTROIDS_DEC_VIEW']):
                    folded_centroids_array[i, :, 1] = np.zeros(folded_centroids_array[i].shape[0])
                if 'EXPLAIN_OG_VIEW' in target_row and not np.isnan(target_row['EXPLAIN_OG_VIEW']):
                    folded_og_array[i] = np.zeros(folded_og_array[i].shape)
            i = i + 1
        filter_channels = np.array([0, 1, 6, 7, 8, 9, 10, 11])
        return [scalar_values[:, [0, 1, 3, 4, 5, 8, 9, 10, 11]], scalar_values[:, [7]],
                star_array, #Only Teff, Rad and Mass
                global_flux_array,
                folded_flux_even_array,
                folded_flux_odd_array,
                folded_flux_even_har_array,
                folded_flux_odd_har_array,
                folded_centroids_array, folded_og_array], \
            batch_data_values



import logging
import os.path
import shutil

from keras.callbacks import ModelCheckpoint
import tensorflow as tf
import numpy as np


def sync_to_numpy_or_python_type(tensors):
    """Took this method from abandonned tensorflow_addons

    Syncs and converts a structure of `Tensor`s to `NumPy` arrays or Python
    scalar types.

    For each tensor, it calls `tensor.numpy()`. If the result is a scalar value,
    it converts it to a Python type, such as a float or int, by calling
    `result.item()`.

    Numpy scalars are converted, as Python types are often more convenient to
    deal with. This is especially useful for bfloat16 Numpy scalars, which don't
    support as many operations as other Numpy values.

    Async strategies (such as `TPUStrategy` and `ParameterServerStrategy`) are
    forced to
    sync during this process.

    Args:
      tensors: A structure of tensors.

    Returns:
      `tensors`, but scalar tensors are converted to Python types and non-scalar
      tensors are converted to Numpy arrays.
    """
    if isinstance(tensors, tf.distribute.experimental.coordinator.RemoteValue):
        tensors = tensors.fetch()

    def _to_single_numpy_or_python_type(t):
        # Don't turn ragged or sparse tensors to NumPy.
        if isinstance(t, tf.Tensor):
            t = t.numpy()
        # Strings, ragged and sparse tensors don't have .item(). Return them
        # as-is.
        if not isinstance(t, (np.ndarray, np.generic)):
            return t
        return t.item() if np.ndim(t) == 0 else t

    return tf.nest.map_structure(_to_single_numpy_or_python_type, tensors)


class ModelCheckpointCallback(ModelCheckpoint):
    def __init__(self, filepath, original_path, monitor: str = "val_loss", verbose: int = 0,
                 save_best_only: bool = False,
                 save_weights_only: bool = False, mode: str = "auto", save_freq="epoch", options=None,
                 initial_value_threshold=None, filter_metrics: list[str] = [], **kwargs):
        super().__init__(filepath, monitor, verbose, save_best_only, save_weights_only, mode, save_freq, options,
                         initial_value_threshold, **kwargs)
        self.original_path = original_path
        self.filter_metrics = filter_metrics

    def on_epoch_end(self, epoch, logs=None):
        self.epochs_since_last_save += 1
        if self.save_freq == "epoch":
            self._save_model(epoch=epoch, batch=None, logs=logs)

    def copy_files(self):
        if os.path.exists(self.original_path + '/training_log.csv'):
            shutil.copyfile(self.original_path + '/training_log.csv', self.filepath + '/training_log.csv')
        if os.path.exists(self.original_path + '/train_dataset.csv'):
            shutil.copyfile(self.original_path + '/train_dataset.csv', self.filepath + '/train_dataset.csv')
        if os.path.exists(self.original_path + '/validation_dataset.csv'):
            shutil.copyfile(self.original_path + '/validation_dataset.csv', self.filepath + '/validation_dataset.csv')
        if os.path.exists(self.original_path + '/network.png'):
            shutil.copyfile(self.original_path + '/network.png', self.filepath + '/network.png')
        if os.path.exists(self.original_path + '/metrics.png'):
            shutil.copyfile(self.original_path + '/metrics.png', self.filepath + '/metrics.png')

    def _save_model(self, epoch, batch, logs):
        """Saves the model.

        Args:
            epoch: the epoch this iteration is in.
            batch: the batch this iteration is in. `None` if the `save_freq`
              is set to `epoch`.
            logs: the `logs` dict passed in to `on_batch_end` or `on_epoch_end`.
        """
        logs = logs or {}

        if (
                isinstance(self.save_freq, int)
                or self.epochs_since_last_save >= self.period
        ):
            # Block only when saving interval is reached.
            logs = sync_to_numpy_or_python_type(logs)
            self.epochs_since_last_save = 0
            filepath = self._get_file_path(epoch, batch, logs)

            try:
                if self.save_best_only:
                    current = logs.get(self.monitor)
                    metrics_for_filter = [logs.get(filter_metric) for filter_metric in self.filter_metrics]
                    if current is None:
                        logging.warning(
                            "Can save best model only with %s available, "
                            "skipping.",
                            self.monitor,
                        )
                    else:
                        if self.monitor_op(current, self.best) and np.all(np.array(metrics_for_filter) > 0):
                            if self.verbose > 0:
                                logging.info(
                                    f"\nEpoch {epoch + 1}: {self.monitor} "
                                    "improved "
                                    f"from {self.best:.5f} to {current:.5f}, "
                                    f"saving model to {filepath}"
                                )
                            self.best = current
                            if self.save_weights_only:
                                self.model.save_weights(
                                    filepath,
                                    overwrite=True,
                                    options=self._options,
                                )
                            else:
                                self.model.save(
                                    filepath,
                                    overwrite=True,
                                    options=self._options,
                                )
                            self.copy_files()
                        else:
                            if self.verbose > 0:
                                logging.info(
                                    f"\nEpoch {epoch + 1}: "
                                    f"{self.monitor} did not improve "
                                    f"from {self.best:.5f}"
                                )
                else:
                    if self.verbose > 0:
                        logging.info(
                            f"\nEpoch {epoch + 1}: saving model to {filepath}"
                        )
                    if self.save_weights_only:
                        self.model.save_weights(
                            filepath, overwrite=True, options=self._options
                        )
                    else:
                        self.model.save(
                            filepath, overwrite=True, options=self._options
                        )
                    self.copy_files()
                self._maybe_remove_file()
            except IsADirectoryError:  # h5py 3.x
                raise IOError(
                    "Please specify a non-directory filepath for "
                    "ModelCheckpoint. Filepath used is an existing "
                    f"directory: {filepath}"
                )
            except IOError as e:  # h5py 2.x
                # `e.errno` appears to be `None` so checking the content of
                # `e.args[0]`.
                if "is a directory" in str(e.args[0]).lower():
                    raise IOError(
                        "Please specify a non-directory filepath for "
                        "ModelCheckpoint. Filepath used is an existing "
                        f"directory: f{filepath}"
                    )
                # Re-throw the error for any other causes.
                raise e



import tensorflow as tf


class RandomReverseLayer(tf.keras.layers.Layer):
    def __init__(self, threshold, axis=-1, **kwargs):
        super(RandomReverseLayer, self).__init__(**kwargs)
        self.axis = axis
        self.threshold = threshold  # The maximum integer threshold

    def call(self, inputs, training=None):
        # Generate a random integer between 0 and threshold
        if training:
            rand_int = tf.random.uniform(shape=(), minval=0, maxval=1)
            return tf.cond(tf.math.less_equal(rand_int, self.threshold),
                           lambda: RandomReverseLayer.reverse_tensor(inputs, [self.axis]),
                           lambda: RandomReverseLayer.return_tensor(inputs))
        else:
            return inputs

    @staticmethod
    def return_tensor(inputs):
        return inputs

    @staticmethod
    def reverse_tensor(inputs, axis):
        return tf.reverse(inputs, axis=axis)

    def get_config(self):
        config = super(RandomReverseLayer, self).get_config()
        config.update({"threshold": self.threshold})
        return config


import tensorflow as tf
from keras import backend as K


def intransit_weighted_mean_squared_error(y_true, y_pred):
    y_pred = tf.convert_to_tensor(y_pred)
    y_true = tf.cast(y_true, y_pred.dtype)
    squared_diff = tf.math.squared_difference(y_pred, y_true)
    it_mask = tf.where(tf.less(y_true, 1))
    oot_mask = tf.where(tf.equal(y_true, 1))
    oot_squared_diff = K.mean(tf.boolean_mask(squared_diff, oot_mask), axis=-1)
    it_squared_diff = K.mean(tf.boolean_mask(squared_diff, it_mask), axis=-1)
    return K.sum(oot_squared_diff, it_squared_diff)
